@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2005-2011 Erik de Castro Lopo <erikd@mega-nerd.com>
+** Copyright (C) 2005-2016 Erik de Castro Lopo <erikd@mega-nerd.com>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU Lesser General Public License as published by
@@ -75,17 +75,19 @@
 */
 
 typedef struct
-{	unsigned char srate [8] ;
-	unsigned int fmt_id ;
-	unsigned int fmt_flags ;
-	unsigned int pkt_bytes ;
-	unsigned int pkt_frames ;
-	unsigned int channels_per_frame ;
-	unsigned int bits_per_chan ;
+{	uint8_t		srate [8] ;
+	uint32_t	fmt_id ;
+	uint32_t	fmt_flags ;
+	uint32_t	pkt_bytes ;
+	uint32_t	frames_per_packet ;
+	uint32_t	channels_per_frame ;
+	uint32_t	bits_per_chan ;
 } DESC_CHUNK ;
 
 typedef struct
-{	int chanmap_tag ;
+{	int			chanmap_tag ;
+
+	ALAC_DECODER_INFO	alac ;
 } CAF_PRIVATE ;
 
 /*------------------------------------------------------------------------------
@@ -95,8 +97,17 @@ typedef struct
 static int	caf_close (SF_PRIVATE *psf) ;
 static int	caf_read_header (SF_PRIVATE *psf) ;
 static int	caf_write_header (SF_PRIVATE *psf, int calc_length) ;
+static int	caf_write_tailer (SF_PRIVATE *psf) ;
 static int	caf_command (SF_PRIVATE *psf, int command, void *data, int datasize) ;
 static int	caf_read_chanmap (SF_PRIVATE * psf, sf_count_t chunk_size) ;
+static int	caf_read_strings (SF_PRIVATE * psf, sf_count_t chunk_size) ;
+static void	caf_write_strings (SF_PRIVATE * psf, int location) ;
+
+
+static int caf_set_chunk (SF_PRIVATE *psf, const SF_CHUNK_INFO * chunk_info) ;
+static SF_CHUNK_ITERATOR * caf_next_chunk_iterator (SF_PRIVATE *psf, SF_CHUNK_ITERATOR * iterator) ;
+static int caf_get_chunk_size (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info) ;
+static int caf_get_chunk_data (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info) ;
 
 /*------------------------------------------------------------------------------
 ** Public function.
@@ -104,18 +115,25 @@ static int	caf_read_chanmap (SF_PRIVATE * psf, sf_count_t chunk_size) ;
 
 int
 caf_open (SF_PRIVATE *psf)
-{	int	subformat, format, error = 0 ;
+{	CAF_PRIVATE * pcaf ;
+	int	subformat, format, error = 0 ;
+
+	if ((psf->container_data = calloc (1, sizeof (CAF_PRIVATE))) == NULL)
+		return SFE_MALLOC_FAILED ;
+
+	pcaf = psf->container_data ;
 
 	if (psf->file.mode == SFM_READ || (psf->file.mode == SFM_RDWR && psf->filelength > 0))
 	{	if ((error = caf_read_header (psf)))
 			return error ;
+
+		psf->next_chunk_iterator = caf_next_chunk_iterator ;
+		psf->get_chunk_size = caf_get_chunk_size ;
+		psf->get_chunk_data = caf_get_chunk_data ;
 		} ;
 
 	subformat = SF_CODEC (psf->sf.format) ;
 
-	if ((psf->container_data = calloc (1, sizeof (CAF_PRIVATE))) == NULL)
-		return SFE_MALLOC_FAILED ;
-
 	if (psf->file.mode == SFM_WRITE || psf->file.mode == SFM_RDWR)
 	{	if (psf->is_pipe)
 			return SFE_NO_PIPE_WRITE ;
@@ -133,7 +151,7 @@ caf_open (SF_PRIVATE *psf)
 			psf->sf.frames = 0 ;
 			} ;
 
-		psf->str_flags = SF_STR_ALLOW_START ;
+		psf->strings.flags = SF_STR_ALLOW_START | SF_STR_ALLOW_END ;
 
 		/*
 		**	By default, add the peak chunk to floating point files. Default behaviour
@@ -148,7 +166,8 @@ caf_open (SF_PRIVATE *psf)
 		if ((error = caf_write_header (psf, SF_FALSE)) != 0)
 			return error ;
 
-		psf->write_header = caf_write_header ;
+		psf->write_header	= caf_write_header ;
+		psf->set_chunk		= caf_set_chunk ;
 		} ;
 
 	psf->container_close = caf_close ;
@@ -178,6 +197,18 @@ caf_open (SF_PRIVATE *psf)
 		case SF_FORMAT_DOUBLE :
 					error = double64_init (psf) ;
 					break ;
+
+		case SF_FORMAT_ALAC_16 :
+		case SF_FORMAT_ALAC_20 :
+		case SF_FORMAT_ALAC_24 :
+		case SF_FORMAT_ALAC_32 :
+					if (psf->file.mode == SFM_READ)
+						/* Only pass the ALAC_DECODER_INFO in read mode. */
+						error = alac_init (psf, &pcaf->alac) ;
+					else
+						error = alac_init (psf, NULL) ;
+					break ;
+
 		/* Lite remove end */
 
 		default :
@@ -191,7 +222,9 @@ static int
 caf_close (SF_PRIVATE *psf)
 {
 	if (psf->file.mode == SFM_WRITE || psf->file.mode == SFM_RDWR)
+	{	caf_write_tailer (psf) ;
 		caf_write_header (psf, SF_TRUE) ;
+		} ;
 
 	return 0 ;
 } /* caf_close */
@@ -220,11 +253,42 @@ caf_command (SF_PRIVATE * psf, int command, void * UNUSED (data), int UNUSED (da
 
 static int
 decode_desc_chunk (SF_PRIVATE *psf, const DESC_CHUNK *desc)
-{	int format ;
+{	int format = SF_FORMAT_CAF ;
 
 	psf->sf.channels = desc->channels_per_frame ;
 
-	format = SF_FORMAT_CAF | (psf->endian == SF_ENDIAN_LITTLE ? SF_ENDIAN_LITTLE : 0) ;
+	if (desc->fmt_id == alac_MARKER)
+	{	CAF_PRIVATE	*pcaf ;
+
+		if ((pcaf = psf->container_data) != NULL)
+		{	switch (desc->fmt_flags)
+			{	case 1 :
+					pcaf->alac.bits_per_sample = 16 ;
+					format |= SF_FORMAT_ALAC_16 ;
+					break ;
+				case 2 :
+					pcaf->alac.bits_per_sample = 20 ;
+					format |= SF_FORMAT_ALAC_20 ;
+					break ;
+				case 3 :
+					pcaf->alac.bits_per_sample = 24 ;
+					format |= SF_FORMAT_ALAC_24 ;
+					break ;
+				case 4 :
+					pcaf->alac.bits_per_sample = 32 ;
+					format |= SF_FORMAT_ALAC_32 ;
+					break ;
+				default :
+					psf_log_printf (psf, "Bad ALAC format flag value of %d\n", desc->fmt_flags) ;
+				} ;
+
+			pcaf->alac.frames_per_packet = desc->frames_per_packet ;
+			} ;
+
+		return format ;
+		} ;
+
+	format |= psf->endian == SF_ENDIAN_LITTLE ? SF_ENDIAN_LITTLE : 0 ;
 
 	if (desc->fmt_id == lpcm_MARKER && desc->fmt_flags & 1)
 	{	/* Floating point data. */
@@ -238,12 +302,7 @@ decode_desc_chunk (SF_PRIVATE *psf, const DESC_CHUNK *desc)
 			} ;
 		} ;
 
-	if ((desc->fmt_flags & 1) != 0)
-	{	psf_log_printf (psf, "**** Ooops, 'desc' chunk suggests float data, but other info invalid.\n") ;
-		return 0 ;
-		} ;
-
-	if (desc->fmt_id == lpcm_MARKER)
+	if (desc->fmt_id == lpcm_MARKER && (desc->fmt_flags & 1) == 0)
 	{	/* Integer data. */
 		if (desc->bits_per_chan == 32 && desc->pkt_bytes == 4 * desc->channels_per_frame)
 		{	psf->bytewidth = 4 ;
@@ -273,17 +332,24 @@ decode_desc_chunk (SF_PRIVATE *psf, const DESC_CHUNK *desc)
 		return format | SF_FORMAT_ULAW ;
 		} ;
 
+	psf_log_printf (psf, "**** Unknown format identifier.\n") ;
+
 	return 0 ;
 } /* decode_desc_chunk */
 
 static int
 caf_read_header (SF_PRIVATE *psf)
-{	DESC_CHUNK desc ;
+{	CAF_PRIVATE	*pcaf ;
+	BUF_UNION	ubuf ;
+	DESC_CHUNK desc ;
 	sf_count_t chunk_size ;
 	double srate ;
 	short version, flags ;
 	int marker, k, have_data = 0, error ;
 
+	if ((pcaf = psf->container_data) == NULL)
+		return SFE_INTERNAL ;
+
 	memset (&desc, 0, sizeof (desc)) ;
 
 	/* Set position to start of file to begin reading header. */
@@ -292,10 +358,10 @@ caf_read_header (SF_PRIVATE *psf)
 	if (marker != caff_MARKER)
 		return SFE_CAF_NOT_CAF ;
 
-	psf_binheader_readf (psf, "mE8b", &marker, &chunk_size, psf->u.ucbuf, 8) ;
-	srate = double64_be_read (psf->u.ucbuf) ;
-	snprintf (psf->u.cbuf, sizeof (psf->u.cbuf), "%5.3f", srate) ;
-	psf_log_printf (psf, "%M : %D\n  Sample rate  : %s\n", marker, chunk_size, psf->u.cbuf) ;
+	psf_binheader_readf (psf, "mE8b", &marker, &chunk_size, ubuf.ucbuf, 8) ;
+	srate = double64_be_read (ubuf.ucbuf) ;
+	snprintf (ubuf.cbuf, sizeof (ubuf.cbuf), "%5.3f", srate) ;
+	psf_log_printf (psf, "%M : %D\n  Sample rate  : %s\n", marker, chunk_size, ubuf.cbuf) ;
 	if (marker != desc_MARKER)
 		return SFE_CAF_NO_DESC ;
 
@@ -306,11 +372,11 @@ caf_read_header (SF_PRIVATE *psf)
 
 	psf->sf.samplerate = lrint (srate) ;
 
-	psf_binheader_readf (psf, "mE44444", &desc.fmt_id, &desc.fmt_flags, &desc.pkt_bytes, &desc.pkt_frames,
+	psf_binheader_readf (psf, "mE44444", &desc.fmt_id, &desc.fmt_flags, &desc.pkt_bytes, &desc.frames_per_packet,
 			&desc.channels_per_frame, &desc.bits_per_chan) ;
 	psf_log_printf (psf, "  Format id    : %M\n  Format flags : %x\n  Bytes / packet   : %u\n"
 			"  Frames / packet  : %u\n  Channels / frame : %u\n  Bits / channel   : %u\n",
-			desc.fmt_id, desc.fmt_flags, desc.pkt_bytes, desc.pkt_frames, desc.channels_per_frame, desc.bits_per_chan) ;
+			desc.fmt_id, desc.fmt_flags, desc.pkt_bytes, desc.frames_per_packet, desc.channels_per_frame, desc.bits_per_chan) ;
 
 	if (desc.channels_per_frame > SF_MAX_CHANNELS)
 	{	psf_log_printf (psf, "**** Bad channels per frame value %u.\n", desc.channels_per_frame) ;
@@ -322,14 +388,28 @@ caf_read_header (SF_PRIVATE *psf)
 
 	psf->sf.channels = desc.channels_per_frame ;
 
-	while (have_data == 0 && psf_ftell (psf) < psf->filelength - SIGNED_SIZEOF (marker))
-	{	psf_binheader_readf (psf, "mE8", &marker, &chunk_size) ;
+	while (1)
+	{	marker = 0 ;
+		chunk_size = 0 ;
+
+		psf_binheader_readf (psf, "mE8", &marker, &chunk_size) ;
+		if (marker == 0)
+		{	sf_count_t pos = psf_ftell (psf) ;
+			psf_log_printf (psf, "Have 0 marker at position %D (0x%x).\n", pos, pos) ;
+			break ;
+			} ;
+		if (chunk_size < 0)
+		{	psf_log_printf (psf, "%M : %D *** Should be >= 0 ***\n", marker, chunk_size) ;
+			break ;
+			} ;
+
+		psf_store_read_chunk_u32 (&psf->rchunks, marker, psf_ftell (psf), chunk_size) ;
 
 		switch (marker)
 		{	case peak_MARKER :
 				psf_log_printf (psf, "%M : %D\n", marker, chunk_size) ;
 				if (chunk_size != CAF_PEAK_CHUNK_SIZE (psf->sf.channels))
-				{	psf_binheader_readf (psf, "j", (int) chunk_size) ;
+				{	psf_binheader_readf (psf, "j", make_size_t (chunk_size)) ;
 					psf_log_printf (psf, "*** File PEAK chunk %D should be %d.\n", chunk_size, CAF_PEAK_CHUNK_SIZE (psf->sf.channels)) ;
 					return SFE_CAF_BAD_PEAK ;
 					} ;
@@ -350,8 +430,8 @@ caf_read_header (SF_PRIVATE *psf)
 					psf->peak_info->peaks [k].value = value ;
 					psf->peak_info->peaks [k].position = position ;
 
-					snprintf (psf->u.cbuf, sizeof (psf->u.cbuf), "    %2d   %-12" PRId64 "   %g\n", k, position, value) ;
-					psf_log_printf (psf, psf->u.cbuf) ;
+					snprintf (ubuf.cbuf, sizeof (ubuf.cbuf), "    %2d   %-12" PRId64 "   %g\n", k, position, value) ;
+					psf_log_printf (psf, ubuf.cbuf) ;
 					} ;
 
 				psf->peak_info->peak_loc = SF_PEAK_START ;
@@ -360,7 +440,7 @@ caf_read_header (SF_PRIVATE *psf)
 			case chan_MARKER :
 				if (chunk_size < 12)
 				{	psf_log_printf (psf, "%M : %D (should be >= 12)\n", marker, chunk_size) ;
-					psf_binheader_readf (psf, "j", (int) chunk_size) ;
+					psf_binheader_readf (psf, "j", make_size_t (chunk_size)) ;
 					break ;
 					}
 
@@ -372,23 +452,101 @@ caf_read_header (SF_PRIVATE *psf)
 
 			case free_MARKER :
 				psf_log_printf (psf, "%M : %D\n", marker, chunk_size) ;
-				psf_binheader_readf (psf, "j", (int) chunk_size) ;
+				psf_binheader_readf (psf, "j", make_size_t (chunk_size)) ;
 				break ;
 
 			case data_MARKER :
-				if (psf->filelength > 0 && chunk_size + psf->headindex != psf->filelength)
-					psf_log_printf (psf, "%M : %D (should be %D)\n", marker, chunk_size, chunk_size + 4) ;
-				else
-					psf_log_printf (psf, "%M : %D\n", marker, chunk_size) ;
 				psf_binheader_readf (psf, "E4", &k) ;
+				if (chunk_size == -1)
+				{	psf_log_printf (psf, "%M : -1\n") ;
+					chunk_size = psf->filelength - psf->headindex ;
+					}
+				else if (psf->filelength > 0 && chunk_size > psf->filelength - psf->headindex + 10)
+				{	psf_log_printf (psf, "%M : %D (should be %D)\n", marker, chunk_size, psf->filelength - psf->headindex - 8) ;
+					psf->datalength = psf->filelength - psf->headindex - 8 ;
+					}
+				else
+				{	psf_log_printf (psf, "%M : %D\n", marker, chunk_size) ;
+					/* Subtract the 4 bytes of the 'edit' field above. */
+					psf->datalength = chunk_size - 4 ;
+					} ;
+
 				psf_log_printf (psf, "  edit : %u\n", k) ;
+
+				psf->dataoffset = psf->headindex ;
+				if (psf->datalength + psf->dataoffset < psf->filelength)
+					psf->dataend = psf->datalength + psf->dataoffset ;
+
+				psf_binheader_readf (psf, "j", make_size_t (psf->datalength)) ;
 				have_data = 1 ;
 				break ;
 
-			default :
-				psf_log_printf (psf, " %M : %D (skipped)\n", marker, chunk_size) ;
-				psf_binheader_readf (psf, "j", (int) chunk_size) ;
+			case kuki_MARKER :
+				psf_log_printf (psf, "%M : %D\n", marker, chunk_size) ;
+				pcaf->alac.kuki_offset = psf_ftell (psf) - 12 ;
+				psf_binheader_readf (psf, "j", make_size_t (chunk_size)) ;
 				break ;
+
+			case pakt_MARKER :
+				if (chunk_size < 24)
+				{	psf_log_printf (psf, "%M : %D (should be > 24)\n", marker, chunk_size) ;
+					return SFE_MALFORMED_FILE ;
+					}
+				else if (chunk_size > psf->filelength - psf->headindex)
+				{	psf_log_printf (psf, "%M : %D (should be < %D)\n", marker, chunk_size, psf->filelength - psf->headindex) ;
+					return SFE_MALFORMED_FILE ;
+					}
+				else
+					psf_log_printf (psf, "%M : %D\n", marker, chunk_size) ;
+
+				psf_binheader_readf (psf, "E8844", &pcaf->alac.packets, &pcaf->alac.valid_frames,
+									&pcaf->alac.priming_frames, &pcaf->alac.remainder_frames) ;
+
+				psf_log_printf (psf,
+						"  Packets          : %D\n"
+						"  Valid frames     : %D\n"
+						"  Priming frames   : %d\n"
+						"  Remainder frames : %d\n",
+						pcaf->alac.packets, pcaf->alac.valid_frames, pcaf->alac.priming_frames,
+						pcaf->alac.remainder_frames
+						) ;
+
+				if (pcaf->alac.packets == 0 && pcaf->alac.valid_frames == 0
+							&& pcaf->alac.priming_frames == 0 && pcaf->alac.remainder_frames == 0)
+					psf_log_printf (psf, "*** 'pakt' chunk header is all zero.\n") ;
+
+				pcaf->alac.pakt_offset = psf_ftell (psf) - 12 ;
+				psf_binheader_readf (psf, "j", make_size_t (chunk_size) - 24) ;
+				break ;
+
+			case info_MARKER :
+				if (chunk_size < 4)
+				{	psf_log_printf (psf, "%M : %D (should be > 4)\n", marker, chunk_size) ;
+					return SFE_MALFORMED_FILE ;
+					}
+				else if (chunk_size > psf->filelength - psf->headindex)
+				{	psf_log_printf (psf, "%M : %D (should be < %D)\n", marker, chunk_size, psf->filelength - psf->headindex) ;
+					return SFE_MALFORMED_FILE ;
+					} ;
+				psf_log_printf (psf, "%M : %D\n", marker, chunk_size) ;
+				caf_read_strings (psf, chunk_size) ;
+				break ;
+
+			default :
+				psf_log_printf (psf, "%M : %D (skipped)\n", marker, chunk_size) ;
+				psf_binheader_readf (psf, "j", make_size_t (chunk_size)) ;
+				break ;
+			} ;
+
+		if (marker != data_MARKER && chunk_size >= 0xffffff00)
+			break ;
+
+		if (! psf->sf.seekable && have_data)
+			break ;
+
+		if (psf_ftell (psf) >= psf->filelength - SIGNED_SIZEOF (chunk_size))
+		{	psf_log_printf (psf, "End\n") ;
+			break ;
 			} ;
 		} ;
 
@@ -397,12 +555,10 @@ caf_read_header (SF_PRIVATE *psf)
 		return SFE_MALFORMED_FILE ;
 		} ;
 
-	psf_log_printf (psf, "End\n") ;
-
-	psf->dataoffset = psf_ftell (psf) ;
-	psf->datalength = psf->filelength - psf->dataoffset ;
 	psf->endian = (desc.fmt_flags & 2) ? SF_ENDIAN_LITTLE : SF_ENDIAN_BIG ;
 
+	psf_fseek (psf, psf->dataoffset, SEEK_SET) ;
+
 	if ((psf->sf.format = decode_desc_chunk (psf, &desc)) == 0)
 		return SFE_UNSUPPORTED_ENCODING ;
 
@@ -417,10 +573,12 @@ caf_read_header (SF_PRIVATE *psf)
 
 static int
 caf_write_header (SF_PRIVATE *psf, int calc_length)
-{	CAF_PRIVATE	*pcaf ;
+{	BUF_UNION	ubuf ;
+	CAF_PRIVATE	*pcaf ;
 	DESC_CHUNK desc ;
 	sf_count_t current, free_len ;
-	int subformat ;
+	uint32_t uk ;
+	int subformat, append_free_block = SF_TRUE ;
 
 	if ((pcaf = psf->container_data) == NULL)
 		return SFE_INTERNAL ;
@@ -452,8 +610,8 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 	/* 'desc' marker and chunk size. */
 	psf_binheader_writef (psf, "Em8", desc_MARKER, (sf_count_t) (sizeof (DESC_CHUNK))) ;
 
- 	double64_be_write (1.0 * psf->sf.samplerate, psf->u.ucbuf) ;
-	psf_binheader_writef (psf, "b", psf->u.ucbuf, make_size_t (8)) ;
+ 	double64_be_write (1.0 * psf->sf.samplerate, ubuf.ucbuf) ;
+	psf_binheader_writef (psf, "b", ubuf.ucbuf, make_size_t (8)) ;
 
 	subformat = SF_CODEC (psf->sf.format) ;
 
@@ -475,7 +633,7 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 			desc.fmt_id = lpcm_MARKER ;
 			psf->bytewidth = 1 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 8 ;
 			break ;
@@ -484,7 +642,7 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 			desc.fmt_id = lpcm_MARKER ;
 			psf->bytewidth = 2 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 16 ;
 			break ;
@@ -492,7 +650,7 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 		case SF_FORMAT_PCM_24 :
 			psf->bytewidth = 3 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 24 ;
 			desc.fmt_id = lpcm_MARKER ;
@@ -502,7 +660,7 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 			desc.fmt_id = lpcm_MARKER ;
 			psf->bytewidth = 4 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 32 ;
 			break ;
@@ -512,7 +670,7 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 			desc.fmt_flags |= 1 ;
 			psf->bytewidth = 4 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 32 ;
 			break ;
@@ -522,7 +680,7 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 			desc.fmt_flags |= 1 ;
 			psf->bytewidth = 8 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 64 ;
 			break ;
@@ -531,7 +689,7 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 			desc.fmt_id = alaw_MARKER ;
 			psf->bytewidth = 1 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 8 ;
 			break ;
@@ -540,21 +698,29 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 			desc.fmt_id = ulaw_MARKER ;
 			psf->bytewidth = 1 ;
 			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
-			desc.pkt_frames = 1 ;
+			desc.frames_per_packet = 1 ;
 			desc.channels_per_frame = psf->sf.channels ;
 			desc.bits_per_chan = 8 ;
 			break ;
 
+		case SF_FORMAT_ALAC_16 :
+		case SF_FORMAT_ALAC_20 :
+		case SF_FORMAT_ALAC_24 :
+		case SF_FORMAT_ALAC_32 :
+			desc.fmt_id = alac_MARKER ;
+			desc.pkt_bytes = psf->bytewidth * psf->sf.channels ;
+			desc.channels_per_frame = psf->sf.channels ;
+			alac_get_desc_chunk_items (subformat, &desc.fmt_flags, &desc.frames_per_packet) ;
+			append_free_block = SF_FALSE ;
+			break ;
+
 		default :
 			return SFE_UNIMPLEMENTED ;
 		} ;
 
-	psf_binheader_writef (psf, "mE44444", desc.fmt_id, desc.fmt_flags, desc.pkt_bytes, desc.pkt_frames, desc.channels_per_frame, desc.bits_per_chan) ;
+	psf_binheader_writef (psf, "mE44444", desc.fmt_id, desc.fmt_flags, desc.pkt_bytes, desc.frames_per_packet, desc.channels_per_frame, desc.bits_per_chan) ;
 
-#if 0
-	if (psf->str_flags & SF_STR_LOCATE_START)
-		caf_write_strings (psf, SF_STR_LOCATE_START) ;
-#endif
+	caf_write_strings (psf, SF_STR_LOCATE_START) ;
 
 	if (psf->peak_info != NULL)
 	{	int k ;
@@ -566,11 +732,17 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 	if (psf->channel_map && pcaf->chanmap_tag)
 		psf_binheader_writef (psf, "Em8444", chan_MARKER, (sf_count_t) 12, pcaf->chanmap_tag, 0, 0) ;
 
-	/* Add free chunk so that the actual audio data starts at a multiple 0x1000. */
-	free_len = 0x1000 - psf->headindex - 16 - 12 ;
-	while (free_len < 0)
-		free_len += 0x1000 ;
-	psf_binheader_writef (psf, "Em8z", free_MARKER, free_len, (int) free_len) ;
+	/* Write custom headers. */
+	for (uk = 0 ; uk < psf->wchunks.used ; uk++)
+		psf_binheader_writef (psf, "m44b", (int) psf->wchunks.chunks [uk].mark32, 0, psf->wchunks.chunks [uk].len, psf->wchunks.chunks [uk].data, make_size_t (psf->wchunks.chunks [uk].len)) ;
+
+	if (append_free_block)
+	{	/* Add free chunk so that the actual audio data starts at a multiple 0x1000. */
+		free_len = 0x1000 - psf->headindex - 16 - 12 ;
+		while (free_len < 0)
+			free_len += 0x1000 ;
+		psf_binheader_writef (psf, "Em8z", free_MARKER, free_len, (int) free_len) ;
+		} ;
 
 	psf_binheader_writef (psf, "Em84", data_MARKER, psf->datalength + 4, 0) ;
 
@@ -587,6 +759,36 @@ caf_write_header (SF_PRIVATE *psf, int calc_length)
 	return psf->error ;
 } /* caf_write_header */
 
+static int
+caf_write_tailer (SF_PRIVATE *psf)
+{
+	/* Reset the current header buffer length to zero. */
+	psf->header [0] = 0 ;
+	psf->headindex = 0 ;
+
+	if (psf->bytewidth > 0 && psf->sf.seekable == SF_TRUE)
+	{	psf->datalength = psf->sf.frames * psf->bytewidth * psf->sf.channels ;
+		psf->dataend = psf->dataoffset + psf->datalength ;
+		} ;
+
+	if (psf->dataend > 0)
+		psf_fseek (psf, psf->dataend, SEEK_SET) ;
+	else
+		psf->dataend = psf_fseek (psf, 0, SEEK_END) ;
+
+	if (psf->dataend & 1)
+		psf_binheader_writef (psf, "z", 1) ;
+
+	if (psf->strings.flags & SF_STR_LOCATE_END)
+		caf_write_strings (psf, SF_STR_LOCATE_END) ;
+
+	/* Write the tailer. */
+	if (psf->headindex > 0)
+		psf_fwrite (psf->header, psf->headindex, 1, psf) ;
+
+	return 0 ;
+} /* caf_write_tailer */
+
 static int
 caf_read_chanmap (SF_PRIVATE * psf, sf_count_t chunk_size)
 {	const AIFF_CAF_CHANNEL_MAP * map_info ;
@@ -604,8 +806,8 @@ caf_read_chanmap (SF_PRIVATE * psf, sf_count_t chunk_size)
 	if (bytesread < chunk_size)
 		psf_binheader_readf (psf, "j", chunk_size - bytesread) ;
 
-	if (map_info->channel_map != NULL)
-	{	size_t chanmap_size = psf->sf.channels * sizeof (psf->channel_map [0]) ;
+	if (map_info && map_info->channel_map != NULL)
+	{	size_t chanmap_size = SF_MIN (psf->sf.channels, layout_tag & 0xff) * sizeof (psf->channel_map [0]) ;
 
 		free (psf->channel_map) ;
 
@@ -618,3 +820,206 @@ caf_read_chanmap (SF_PRIVATE * psf, sf_count_t chunk_size)
 	return 0 ;
 } /* caf_read_chanmap */
 
+
+static uint32_t
+string_hash32 (const char * str)
+{	uint32_t hash = 0x87654321 ;
+
+	while (str [0])
+	{	hash = hash * 333 + str [0] ;
+		str ++ ;
+		} ;
+
+	return hash ;
+} /* string_hash32 */
+
+static int
+caf_read_strings (SF_PRIVATE * psf, sf_count_t chunk_size)
+{	char buf [chunk_size - 4] ;
+	char *key, *value ;
+	uint32_t count, hash ;
+
+	psf_binheader_readf (psf, "E4b", &count, buf, make_size_t (chunk_size - 4)) ;
+	psf_log_printf (psf, " count: %u\n", count) ;
+
+	/* Force terminate `buf` to make sure. */
+	buf [sizeof (buf) - 1] = 0 ;
+
+	for (key = buf ; key < buf + sizeof (buf) ; )
+	{	value = key + strlen (key) + 1 ;
+		if (value > buf + sizeof (buf))
+			break ;
+		psf_log_printf (psf, "   %-12s : %s\n", key, value) ;
+
+		hash = string_hash32 (key) ;
+		switch (hash)
+		{	case 0xC4861943 : /* 'title' */
+				psf_store_string (psf, SF_STR_TITLE, value) ;
+				break ;
+			case 0xAD47A394 : /* 'software' */
+				psf_store_string (psf, SF_STR_SOFTWARE, value) ;
+				break ;
+			case 0x5D178E2A : /* 'copyright' */
+				psf_store_string (psf, SF_STR_COPYRIGHT, value) ;
+				break ;
+			case 0x60E4D0C8 : /* 'artist' */
+				psf_store_string (psf, SF_STR_ARTIST, value) ;
+				break ;
+			case 0x83B5D16A : /* 'genre' */
+				psf_store_string (psf, SF_STR_GENRE, value) ;
+				break ;
+			case 0x15E5FC88 : /* 'comment' */
+			case 0x7C297D5B : /* 'comments' */
+				psf_store_string (psf, SF_STR_COMMENT, value) ;
+				break ;
+			case 0x24A7C347 : /* 'tracknumber' */
+				psf_store_string (psf, SF_STR_TRACKNUMBER, value) ;
+				break ;
+			case 0x50A31EB7 : /* 'date' */
+				psf_store_string (psf, SF_STR_DATE, value) ;
+				break ;
+			case 0x6583545A : /* 'album' */
+				psf_store_string (psf, SF_STR_ALBUM, value) ;
+				break ;
+			case 0xE7C64B6C : /* 'license' */
+				psf_store_string (psf, SF_STR_LICENSE, value) ;
+				break ;
+			default :
+				psf_log_printf (psf, " Unhandled hash 0x%x : /* '%s' */\n", hash, key) ;
+				break ;
+			} ;
+
+		key = value + strlen (value) + 1 ;
+		} ;
+
+	return 0 ;
+} /* caf_read_strings */
+
+struct put_buffer
+{	uint32_t index ;
+	char s [16 * 1024] ;
+} ;
+
+static uint32_t
+put_key_value (struct put_buffer * buf, const char * key, const char * value)
+{	uint32_t written ;
+
+	if (buf->index + strlen (key) + strlen (value) + 2 > sizeof (buf->s))
+		return 0 ;
+
+	written = snprintf (buf->s + buf->index, sizeof (buf->s) - buf->index, "%s%c%s%c", key, 0, value, 0) ;
+
+	if (buf->index + written >= sizeof (buf->s))
+		return 0 ;
+
+	buf->index += written ;
+	return 1 ;
+} /* put_key_value */
+
+static void
+caf_write_strings (SF_PRIVATE * psf, int location)
+{	struct put_buffer buf ;
+ 	const char * cptr ;
+	uint32_t k, string_count = 0 ;
+
+	memset (&buf, 0, sizeof (buf)) ;
+
+	for (k = 0 ; k < SF_MAX_STRINGS ; k++)
+	{	if (psf->strings.data [k].type == 0)
+			break ;
+
+		if (psf->strings.data [k].flags != location)
+			continue ;
+
+		if ((cptr = psf_get_string (psf, psf->strings.data [k].type)) == NULL)
+			continue ;
+
+		switch (psf->strings.data [k].type)
+		{	case SF_STR_TITLE :
+				string_count += put_key_value (&buf, "title", cptr) ;
+				break ;
+			case SF_STR_COPYRIGHT :
+				string_count += put_key_value (&buf, "copyright", cptr) ;
+				break ;
+			case SF_STR_SOFTWARE :
+				string_count += put_key_value (&buf, "software", cptr) ;
+				break ;
+			case SF_STR_ARTIST :
+				string_count += put_key_value (&buf, "artist", cptr) ;
+				break ;
+			case SF_STR_COMMENT :
+				string_count += put_key_value (&buf, "comment", cptr) ;
+				break ;
+			case SF_STR_DATE :
+				string_count += put_key_value (&buf, "date", cptr) ;
+				break ;
+			case SF_STR_ALBUM :
+				string_count += put_key_value (&buf, "album", cptr) ;
+				break ;
+			case SF_STR_LICENSE :
+				string_count += put_key_value (&buf, "license", cptr) ;
+				break ;
+			case SF_STR_TRACKNUMBER :
+				string_count += put_key_value (&buf, "tracknumber", cptr) ;
+				break ;
+			case SF_STR_GENRE :
+				string_count += put_key_value (&buf, "genre", cptr) ;
+				break ;
+
+			default :
+				break ;
+			} ;
+		} ;
+
+	if (string_count == 0 || buf.index == 0)
+		return ;
+
+	psf_binheader_writef (psf, "Em84b", info_MARKER, make_size_8 (buf.index + 4), string_count, buf.s, make_size_t (buf.index)) ;
+} /* caf_write_strings */
+
+/*==============================================================================
+*/
+
+static int
+caf_set_chunk (SF_PRIVATE *psf, const SF_CHUNK_INFO * chunk_info)
+{	return psf_save_write_chunk (&psf->wchunks, chunk_info) ;
+} /* caf_set_chunk */
+
+static SF_CHUNK_ITERATOR *
+caf_next_chunk_iterator (SF_PRIVATE *psf, SF_CHUNK_ITERATOR * iterator)
+{	return psf_next_chunk_iterator (&psf->rchunks, iterator) ;
+} /* caf_next_chunk_iterator */
+
+static int
+caf_get_chunk_size (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info)
+{	int indx ;
+
+	if ((indx = psf_find_read_chunk_iterator (&psf->rchunks, iterator)) < 0)
+		return SFE_UNKNOWN_CHUNK ;
+
+	chunk_info->datalen = psf->rchunks.chunks [indx].len ;
+
+	return SFE_NO_ERROR ;
+} /* caf_get_chunk_size */
+
+static int
+caf_get_chunk_data (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info)
+{	int indx ;
+	sf_count_t pos ;
+
+	if ((indx = psf_find_read_chunk_iterator (&psf->rchunks, iterator)) < 0)
+		return SFE_UNKNOWN_CHUNK ;
+
+	if (chunk_info->data == NULL)
+		return SFE_BAD_CHUNK_DATA_PTR ;
+
+	chunk_info->id_size = psf->rchunks.chunks [indx].id_size ;
+	memcpy (chunk_info->id, psf->rchunks.chunks [indx].id, sizeof (chunk_info->id) / sizeof (*chunk_info->id)) ;
+
+	pos = psf_ftell (psf) ;
+	psf_fseek (psf, psf->rchunks.chunks [indx].offset, SEEK_SET) ;
+	psf_fread (chunk_info->data, SF_MIN (chunk_info->datalen, psf->rchunks.chunks [indx].len), 1, psf) ;
+	psf_fseek (psf, pos, SEEK_SET) ;
+
+	return SFE_NO_ERROR ;
+} /* caf_get_chunk_data */
