@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 1999-2011 Erik de Castro Lopo <erikd@mega-nerd.com>
+** Copyright (C) 1999-2016 Erik de Castro Lopo <erikd@mega-nerd.com>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU Lesser General Public License as published by
@@ -16,43 +16,47 @@
 ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include	<config.h>
+#include <config.h>
 
-#include	<stdarg.h>
-#include	<string.h>
-#include	<ctype.h>
-#include	<math.h>
-#include	<time.h>
-#include	<sys/time.h>
-
-#include	"sndfile.h"
-#include	"sfendian.h"
-#include	"common.h"
+#include <stdarg.h>
+#include <string.h>
+#if HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+#include <ctype.h>
+#include <math.h>
+#include <time.h>
+#if HAVE_SYS_TIME_H
+#include <sys/time.h>
+#endif
+#include "sndfile.h"
+#include "sfendian.h"
+#include "common.h"
 
 /*-----------------------------------------------------------------------------------------------
-** psf_log_printf allows libsndfile internal functions to print to an internal logbuffer which
+** psf_log_printf allows libsndfile internal functions to print to an internal parselog which
 ** can later be displayed.
 ** The format specifiers are as for printf but without the field width and other modifiers.
-** Printing is performed to the logbuffer char array of the SF_PRIVATE struct.
+** Printing is performed to the parselog char array of the SF_PRIVATE struct.
 ** Printing is done in such a way as to guarantee that the log never overflows the end of the
-** logbuffer array.
+** parselog array.
 */
 
 static inline void
 log_putchar (SF_PRIVATE *psf, char ch)
-{	if (psf->logindex < SIGNED_SIZEOF (psf->logbuffer) - 1)
-	{	psf->logbuffer [psf->logindex++] = ch ;
-		psf->logbuffer [psf->logindex] = 0 ;
+{	if (psf->parselog.indx < SIGNED_SIZEOF (psf->parselog.buf) - 1)
+	{	psf->parselog.buf [psf->parselog.indx++] = ch ;
+		psf->parselog.buf [psf->parselog.indx] = 0 ;
 		} ;
 	return ;
 } /* log_putchar */
 
 void
 psf_log_printf (SF_PRIVATE *psf, const char *format, ...)
-{	va_list			ap ;
-	unsigned int	u ;
-	int				d, tens, shift, width, width_specifier, left_align ;
-	char			c, *strptr, istr [5], lead_char, sign_char ;
+{	va_list		ap ;
+	uint32_t	u ;
+	int			d, tens, shift, width, width_specifier, left_align, slen ;
+	char		c, *strptr, istr [5], lead_char, sign_char ;
 
 	va_start (ap, format) ;
 
@@ -109,7 +113,8 @@ psf_log_printf (SF_PRIVATE *psf, const char *format, ...)
 					strptr = va_arg (ap, char *) ;
 					if (strptr == NULL)
 						break ;
-					width_specifier -= strlen (strptr) ;
+					slen = strlen (strptr) ;
+					width_specifier = width_specifier >= slen ? width_specifier - slen : 0 ;
 					if (left_align == SF_FALSE)
 						while (width_specifier -- > 0)
 							log_putchar (psf, ' ') ;
@@ -266,7 +271,7 @@ psf_log_printf (SF_PRIVATE *psf, const char *format, ...)
 						} ;
 					shift = 28 ;
 					width = (width_specifier < 8) ? 8 : width_specifier ;
-					while (! ((0xF << shift) & d))
+					while (! ((((uint32_t) 0xF) << shift) & d))
 					{	shift -= 4 ;
 						width -- ;
 						} ;
@@ -755,38 +760,38 @@ psf_binheader_writef (SF_PRIVATE *psf, const char *format, ...)
 */
 
 #if (CPU_IS_BIG_ENDIAN == 1)
-#define	GET_MARKER(ptr)	(	((ptr) [0] << 24)	| ((ptr) [1] << 16) |	\
-							((ptr) [2] << 8)	| ((ptr) [3]) )
+#define	GET_MARKER(ptr)	(	(((uint32_t) (ptr) [0]) << 24)	| ((ptr) [1] << 16) |	\
+							((ptr) [2] << 8)	| ((ptr) [3]))
 
 #elif (CPU_IS_LITTLE_ENDIAN == 1)
 #define	GET_MARKER(ptr)	(	((ptr) [0])			| ((ptr) [1] << 8) |	\
-							((ptr) [2] << 16)	| ((ptr) [3] << 24) )
+							((ptr) [2] << 16)	| (((uint32_t) (ptr) [3]) << 24))
 
 #else
 #	error "Cannot determine endian-ness of processor."
 #endif
 
-#define	GET_LE_SHORT(ptr)	( ((ptr) [1] << 8) | ((ptr) [0]) )
-#define	GET_BE_SHORT(ptr)	( ((ptr) [0] << 8) | ((ptr) [1]) )
+#define	GET_LE_SHORT(ptr)	(((ptr) [1] << 8) | ((ptr) [0]))
+#define	GET_BE_SHORT(ptr)	(((ptr) [0] << 8) | ((ptr) [1]))
 
-#define	GET_LE_3BYTE(ptr)	( 	((ptr) [2] << 16) | ((ptr) [1] << 8) | ((ptr) [0]) )
-#define	GET_BE_3BYTE(ptr)	( 	((ptr) [0] << 16) | ((ptr) [1] << 8) | ((ptr) [2]) )
+#define	GET_LE_3BYTE(ptr)	(	((ptr) [2] << 16) | ((ptr) [1] << 8) | ((ptr) [0]))
+#define	GET_BE_3BYTE(ptr)	(	((ptr) [0] << 16) | ((ptr) [1] << 8) | ((ptr) [2]))
 
-#define	GET_LE_INT(ptr)		( 	((ptr) [3] << 24)	| ((ptr) [2] << 16) |	\
-								((ptr) [1] << 8)	| ((ptr) [0]) )
+#define	GET_LE_INT(ptr)		(	((ptr) [3] << 24)	| ((ptr) [2] << 16) |	\
+								((ptr) [1] << 8)	| ((ptr) [0]))
 
-#define	GET_BE_INT(ptr)		( 	((ptr) [0] << 24)	| ((ptr) [1] << 16) |	\
-							 	((ptr) [2] << 8)	| ((ptr) [3]) )
+#define	GET_BE_INT(ptr)		(	((ptr) [0] << 24)	| ((ptr) [1] << 16) |	\
+								((ptr) [2] << 8)	| ((ptr) [3]))
 
-#define	GET_LE_8BYTE(ptr)	( 	(((sf_count_t) (ptr) [7]) << 56) | (((sf_count_t) (ptr) [6]) << 48) |	\
-							 	(((sf_count_t) (ptr) [5]) << 40) | (((sf_count_t) (ptr) [4]) << 32) |	\
-							 	(((sf_count_t) (ptr) [3]) << 24) | (((sf_count_t) (ptr) [2]) << 16) |	\
-							 	(((sf_count_t) (ptr) [1]) << 8 ) | ((ptr) [0]))
+#define	GET_LE_8BYTE(ptr)	(	(((sf_count_t) (ptr) [7]) << 56)	| (((sf_count_t) (ptr) [6]) << 48) |	\
+								(((sf_count_t) (ptr) [5]) << 40)	| (((sf_count_t) (ptr) [4]) << 32) |	\
+								(((sf_count_t) (ptr) [3]) << 24)	| (((sf_count_t) (ptr) [2]) << 16) |	\
+								(((sf_count_t) (ptr) [1]) << 8)		| ((ptr) [0]))
 
-#define	GET_BE_8BYTE(ptr)	( 	(((sf_count_t) (ptr) [0]) << 56) | (((sf_count_t) (ptr) [1]) << 48) |	\
-							 	(((sf_count_t) (ptr) [2]) << 40) | (((sf_count_t) (ptr) [3]) << 32) |	\
-							 	(((sf_count_t) (ptr) [4]) << 24) | (((sf_count_t) (ptr) [5]) << 16) |	\
-							 	(((sf_count_t) (ptr) [6]) << 8 ) | ((ptr) [7]))
+#define	GET_BE_8BYTE(ptr)	(	(((sf_count_t) (ptr) [0]) << 56)	| (((sf_count_t) (ptr) [1]) << 48) |	\
+								(((sf_count_t) (ptr) [2]) << 40)	| (((sf_count_t) (ptr) [3]) << 32) |	\
+								(((sf_count_t) (ptr) [4]) << 24)	| (((sf_count_t) (ptr) [5]) << 16) |	\
+								(((sf_count_t) (ptr) [6]) << 8)		| ((ptr) [7]))
 
 
 
@@ -795,21 +800,16 @@ header_read (SF_PRIVATE *psf, void *ptr, int bytes)
 {	int count = 0 ;
 
 	if (psf->headindex >= SIGNED_SIZEOF (psf->header))
-	{	memset (ptr, 0, SIGNED_SIZEOF (psf->header) - psf->headindex) ;
-
-		/* This is the best that we can do. */
-		psf_fseek (psf, bytes, SEEK_CUR) ;
-		return bytes ;
-		} ;
+		return psf_fread (ptr, 1, bytes, psf) ;
 
 	if (psf->headindex + bytes > SIGNED_SIZEOF (psf->header))
 	{	int most ;
 
-		most = SIGNED_SIZEOF (psf->header) - psf->headindex ;
+		most = SIGNED_SIZEOF (psf->header) - psf->headend ;
 		psf_fread (psf->header + psf->headend, 1, most, psf) ;
-		memset ((char *) ptr + most, 0, bytes - most) ;
-
-		psf_fseek (psf, bytes - most, SEEK_CUR) ;
+		memcpy (ptr, psf->header + psf->headend, most) ;
+		psf->headend = psf->headindex += most ;
+		psf_fread ((char *) ptr + most, bytes - most, 1, psf) ;
 		return bytes ;
 		} ;
 
@@ -817,7 +817,7 @@ header_read (SF_PRIVATE *psf, void *ptr, int bytes)
 	{	count = psf_fread (psf->header + psf->headend, 1, bytes - (psf->headend - psf->headindex), psf) ;
 		if (count != bytes - (int) (psf->headend - psf->headindex))
 		{	psf_log_printf (psf, "Error : psf_fread returned short count.\n") ;
-			return 0 ;
+			return count ;
 			} ;
 		psf->headend += count ;
 		} ;
@@ -831,7 +831,6 @@ header_read (SF_PRIVATE *psf, void *ptr, int bytes)
 static void
 header_seek (SF_PRIVATE *psf, sf_count_t position, int whence)
 {
-
 	switch (whence)
 	{	case SEEK_SET :
 			if (position > SIGNED_SIZEOF (psf->header))
@@ -880,8 +879,7 @@ header_seek (SF_PRIVATE *psf, sf_count_t position, int whence)
 
 static int
 header_gets (SF_PRIVATE *psf, char *ptr, int bufsize)
-{
-	int		k ;
+{	int		k ;
 
 	for (k = 0 ; k < bufsize - 1 ; k++)
 	{	if (psf->headindex < psf->headend)
@@ -931,7 +929,7 @@ psf_binheader_readf (SF_PRIVATE *psf, char const *format, ...)
 					psf->rwf_endian = SF_ENDIAN_BIG ;
 					break ;
 
-			case 'm' :
+			case 'm' : /* 4 byte marker value eg 'RIFF' */
 					intptr = va_arg (argptr, unsigned int*) ;
 					ucptr = (unsigned char*) intptr ;
 					byte_count += header_read (psf, ucptr, sizeof (int)) ;
@@ -956,7 +954,7 @@ psf_binheader_readf (SF_PRIVATE *psf, char const *format, ...)
 					byte_count += header_read (psf, charptr, sizeof (char)) ;
 					break ;
 
-			case '2' :
+			case '2' : /* 2 byte value with the current endian-ness */
 					shortptr = va_arg (argptr, unsigned short*) ;
 					*shortptr = 0 ;
 					ucptr = (unsigned char*) shortptr ;
@@ -967,7 +965,7 @@ psf_binheader_readf (SF_PRIVATE *psf, char const *format, ...)
 						*shortptr = GET_LE_SHORT (ucptr) ;
 					break ;
 
-			case '3' :
+			case '3' : /* 3 byte value with the current endian-ness */
 					intptr = va_arg (argptr, unsigned int*) ;
 					*intptr = 0 ;
 					byte_count += header_read (psf, sixteen_bytes, 3) ;
@@ -977,25 +975,25 @@ psf_binheader_readf (SF_PRIVATE *psf, char const *format, ...)
 						*intptr = GET_LE_3BYTE (sixteen_bytes) ;
 					break ;
 
-			case '4' :
+			case '4' : /* 4 byte value with the current endian-ness */
 					intptr = va_arg (argptr, unsigned int*) ;
 					*intptr = 0 ;
 					ucptr = (unsigned char*) intptr ;
 					byte_count += header_read (psf, ucptr, sizeof (int)) ;
 					if (psf->rwf_endian == SF_ENDIAN_BIG)
-						*intptr = GET_BE_INT (ucptr) ;
+						*intptr = psf_get_be32 (ucptr, 0) ;
 					else
-						*intptr = GET_LE_INT (ucptr) ;
+						*intptr = psf_get_le32 (ucptr, 0) ;
 					break ;
 
-			case '8' :
+			case '8' : /* 8 byte value with the current endian-ness */
 					countptr = va_arg (argptr, sf_count_t *) ;
 					*countptr = 0 ;
 					byte_count += header_read (psf, sixteen_bytes, 8) ;
 					if (psf->rwf_endian == SF_ENDIAN_BIG)
-						countdata = GET_BE_8BYTE (sixteen_bytes) ;
+						countdata = psf_get_be64 (sixteen_bytes, 0) ;
 					else
-						countdata = GET_LE_8BYTE (sixteen_bytes) ;
+						countdata = psf_get_le64 (sixteen_bytes, 0) ;
 					*countptr = countdata ;
 					break ;
 
@@ -1025,14 +1023,14 @@ psf_binheader_readf (SF_PRIVATE *psf, char const *format, ...)
 					strptr = va_arg (argptr, char *) ;
 					size   = strlen (strptr) + 1 ;
 					size  += (size & 1) ;
-					longdata = H2LE_INT (size) ;
+					longdata = H2LE_32 (size) ;
 					get_int (psf, longdata) ;
 					memcpy (&(psf->header [psf->headindex]), strptr, size) ;
 					psf->headindex += size ;
 					*/
 					break ;
 
-			case 'b' :
+			case 'b' : /* Raw bytes */
 					charptr = va_arg (argptr, char*) ;
 					count = va_arg (argptr, size_t) ;
 					if (count > 0)
@@ -1068,8 +1066,10 @@ psf_binheader_readf (SF_PRIVATE *psf, char const *format, ...)
 			case 'j' :
 					/* Get the seek position first. */
 					count = va_arg (argptr, size_t) ;
-					header_seek (psf, count, SEEK_CUR) ;
-					byte_count += count ;
+					if (count)
+					{	header_seek (psf, count, SEEK_CUR) ;
+						byte_count += count ;
+						} ;
 					break ;
 
 			default :
@@ -1150,7 +1150,10 @@ psf_log_SF_INFO (SF_PRIVATE *psf)
 {	psf_log_printf (psf, "---------------------------------\n") ;
 
 	psf_log_printf (psf, " Sample rate :   %d\n", psf->sf.samplerate) ;
-	psf_log_printf (psf, " Frames      :   %D\n", psf->sf.frames) ;
+	if (psf->sf.frames == SF_COUNT_MAX)
+		psf_log_printf (psf, " Frames      :   unknown\n") ;
+	else
+		psf_log_printf (psf, " Frames      :   %D\n", psf->sf.frames) ;
 	psf_log_printf (psf, " Channels    :   %d\n", psf->sf.channels) ;
 
 	psf_log_printf (psf, " Format      :   0x%X\n", psf->sf.format) ;
@@ -1182,6 +1185,49 @@ psf_memset (void *s, int c, sf_count_t len)
 	return s ;
 } /* psf_memset */
 
+
+/*
+** Clang refuses to do sizeof (SF_CUES_VAR (cue_count)) so we have to manually
+** bodgy something up instead.
+*/
+
+typedef SF_CUES_VAR (1) SF_CUES_1 ;
+typedef SF_CUES_VAR (2) SF_CUES_2 ;
+
+#define SF_CUES_VAR_SIZE(count)	(sizeof (SF_CUES_1) + count * (sizeof (SF_CUES_2) - sizeof (SF_CUES_1)))
+
+SF_CUES *
+psf_cues_alloc (uint32_t cue_count)
+{	SF_CUES *pcues = calloc (1, SF_CUES_VAR_SIZE (cue_count)) ;
+
+	pcues->cue_count = cue_count ;
+	return pcues ;
+} /* psf_cues_alloc */
+
+SF_CUES *
+psf_cues_dup (const void * ptr)
+{	const SF_CUES *pcues = ptr ;
+	SF_CUES *pnew = psf_cues_alloc (pcues->cue_count) ;
+
+	memcpy (pnew, pcues, SF_CUES_VAR_SIZE (pcues->cue_count)) ;
+	return pnew ;
+} /* psf_cues_dup */
+
+void
+psf_get_cues (SF_PRIVATE * psf, void * data, size_t datasize)
+{
+	if (psf->cues)
+	{	uint32_t cue_count = (datasize - sizeof (uint32_t)) / sizeof (SF_CUE_POINT) ;
+
+		cue_count = SF_MIN (cue_count, psf->cues->cue_count) ;
+		memcpy (data, psf->cues, SF_CUES_VAR_SIZE (cue_count)) ;
+		((SF_CUES*) data)->cue_count = cue_count ;
+		} ;
+
+	return ;
+} /* psf_get_cues */
+
+
 SF_INSTRUMENT *
 psf_instrument_alloc (void)
 {	SF_INSTRUMENT *instr ;
@@ -1293,10 +1339,10 @@ u_bitwidth_to_subformat (int bits)
 
 int32_t
 psf_rand_int32 (void)
-{	static int32_t value = -1 ;
+{	static uint64_t value = 0 ;
 	int k, count ;
 
-	if (value == -1)
+	if (value == 0)
 	{
 #if HAVE_GETTIMEOFDAY
 		struct timeval tv ;
@@ -1309,9 +1355,9 @@ psf_rand_int32 (void)
 
 	count = 4 + (value & 7) ;
 	for (k = 0 ; k < count ; k++)
-		value = 11117 * value + 211231 ;
+		value = (11117 * value + 211231) & 0x7fffffff ;
 
-	return value ;
+	return (int32_t) value ;
 } /* psf_rand_int32 */
 
 void
@@ -1365,6 +1411,28 @@ psf_strlcpy_crlf (char *dest, const char *src, size_t destmax, size_t srcmax)
 	*dest = 0 ;
 } /* psf_strlcpy_crlf */
 
+sf_count_t
+psf_decode_frame_count (SF_PRIVATE *psf)
+{	sf_count_t count, readlen, total = 0 ;
+	BUF_UNION	ubuf ;
+
+	/* If we're reading from a pipe or the file is too long, just return SF_COUNT_MAX. */
+	if (psf_is_pipe (psf) || psf->datalength > 0x1000000)
+		return SF_COUNT_MAX ;
+
+	psf_fseek (psf, psf->dataoffset, SEEK_SET) ;
+
+	readlen = ARRAY_LEN (ubuf.ibuf) / psf->sf.channels ;
+	readlen *= psf->sf.channels ;
+
+	while ((count = psf->read_int (psf, ubuf.ibuf, readlen)) > 0)
+		total += count ;
+
+	psf_fseek (psf, psf->dataoffset, SEEK_SET) ;
+
+	return total / psf->sf.channels ;
+} /* psf_decode_frame_count */
+
 /*==============================================================================
 */
 
@@ -1463,3 +1531,169 @@ str_of_endianness (int end)
 	/* Zero length string for SF_ENDIAN_FILE. */
 	return "" ;
 } /* str_of_endianness */
+
+/*==============================================================================
+*/
+
+void
+psf_f2s_array (const float *src, short *dest, int count, int normalize)
+{	float 			normfact ;
+
+	normfact = normalize ? (1.0 * 0x7FFF) : 1.0 ;
+	while (--count >= 0)
+		dest [count] = lrintf (src [count] * normfact) ;
+
+	return ;
+} /* psf_f2s_array */
+
+void
+psf_f2s_clip_array (const float *src, short *dest, int count, int normalize)
+{	float			normfact, scaled_value ;
+
+	normfact = normalize ? (1.0 * 0x8000) : 1.0 ;
+
+	while (--count >= 0)
+	{	scaled_value = src [count] * normfact ;
+		if (CPU_CLIPS_POSITIVE == 0 && scaled_value >= (1.0 * 0x7FFF))
+		{	dest [count] = 0x7FFF ;
+			continue ;
+			} ;
+		if (CPU_CLIPS_NEGATIVE == 0 && scaled_value <= (-8.0 * 0x1000))
+		{	dest [count] = 0x8000 ;
+			continue ;
+			} ;
+
+		dest [count] = lrintf (scaled_value) ;
+		} ;
+
+	return ;
+} /* psf_f2s_clip_array */
+
+void
+psf_d2s_array (const double *src, short *dest, int count, int normalize)
+{	double 			normfact ;
+
+	normfact = normalize ? (1.0 * 0x7FFF) : 1.0 ;
+	while (--count >= 0)
+		dest [count] = lrint (src [count] * normfact) ;
+
+	return ;
+} /* psf_f2s_array */
+
+void
+psf_d2s_clip_array (const double *src, short *dest, int count, int normalize)
+{	double			normfact, scaled_value ;
+
+	normfact = normalize ? (1.0 * 0x8000) : 1.0 ;
+
+	while (--count >= 0)
+	{	scaled_value = src [count] * normfact ;
+		if (CPU_CLIPS_POSITIVE == 0 && scaled_value >= (1.0 * 0x7FFF))
+		{	dest [count] = 0x7FFF ;
+			continue ;
+			} ;
+		if (CPU_CLIPS_NEGATIVE == 0 && scaled_value <= (-8.0 * 0x1000))
+		{	dest [count] = 0x8000 ;
+			continue ;
+			} ;
+
+		dest [count] = lrint (scaled_value) ;
+		} ;
+
+	return ;
+} /* psf_d2s_clip_array */
+
+
+void
+psf_f2i_array (const float *src, int *dest, int count, int normalize)
+{	float 			normfact ;
+
+	normfact = normalize ? (1.0 * 0x7FFFFFFF) : 1.0 ;
+	while (--count >= 0)
+		dest [count] = lrintf (src [count] * normfact) ;
+
+	return ;
+} /* psf_f2i_array */
+
+void
+psf_f2i_clip_array (const float *src, int *dest, int count, int normalize)
+{	float			normfact, scaled_value ;
+
+	normfact = normalize ? (8.0 * 0x10000000) : 1.0 ;
+
+	while (--count >= 0)
+	{	scaled_value = src [count] * normfact ;
+		if (CPU_CLIPS_POSITIVE == 0 && scaled_value >= (1.0 * 0x7FFFFFFF))
+		{	dest [count] = 0x7FFFFFFF ;
+			continue ;
+			} ;
+		if (CPU_CLIPS_NEGATIVE == 0 && scaled_value <= (-8.0 * 0x10000000))
+		{	dest [count] = 0x80000000 ;
+			continue ;
+			} ;
+
+		dest [count] = lrintf (scaled_value) ;
+		} ;
+
+	return ;
+} /* psf_f2i_clip_array */
+
+void
+psf_d2i_array (const double *src, int *dest, int count, int normalize)
+{	double 			normfact ;
+
+	normfact = normalize ? (1.0 * 0x7FFFFFFF) : 1.0 ;
+	while (--count >= 0)
+		dest [count] = lrint (src [count] * normfact) ;
+
+	return ;
+} /* psf_f2i_array */
+
+void
+psf_d2i_clip_array (const double *src, int *dest, int count, int normalize)
+{	double			normfact, scaled_value ;
+
+	normfact = normalize ? (8.0 * 0x10000000) : 1.0 ;
+
+	while (--count >= 0)
+	{	scaled_value = src [count] * normfact ;
+		if (CPU_CLIPS_POSITIVE == 0 && scaled_value >= (1.0 * 0x7FFFFFFF))
+		{	dest [count] = 0x7FFFFFFF ;
+			continue ;
+			} ;
+		if (CPU_CLIPS_NEGATIVE == 0 && scaled_value <= (-8.0 * 0x10000000))
+		{	dest [count] = 0x80000000 ;
+			continue ;
+			} ;
+
+		dest [count] = lrint (scaled_value) ;
+		} ;
+
+	return ;
+} /* psf_d2i_clip_array */
+
+FILE *
+psf_open_tmpfile (char * fname, size_t fnamelen)
+{	const char * tmpdir ;
+	FILE * file ;
+
+	if (OS_IS_WIN32)
+		tmpdir = getenv ("TEMP") ;
+	else
+	{	tmpdir = getenv ("TMPDIR") ;
+		tmpdir = tmpdir == NULL ? "/tmp" : tmpdir ;
+		} ;
+
+	if (tmpdir && access (tmpdir, R_OK | W_OK | X_OK) == 0)
+	{	snprintf (fname, fnamelen, "%s/%x%x-alac.tmp", tmpdir, psf_rand_int32 (), psf_rand_int32 ()) ;
+		if ((file = fopen (fname, "wb+")) != NULL)
+			return file ;
+		} ;
+
+	snprintf (fname, fnamelen, "%x%x-alac.tmp", psf_rand_int32 (), psf_rand_int32 ()) ;
+	if ((file = fopen (fname, "wb+")) != NULL)
+		return file ;
+
+	memset (fname, 0, fnamelen) ;
+	return NULL ;
+} /* psf_open_tmpfile */
