@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2002-2011 Erik de Castro Lopo <erikd@mega-nerd.com>
+** Copyright (C) 2002-2016 Erik de Castro Lopo <erikd@mega-nerd.com>
 ** Copyright (C) 2002-2005 Michael Smith <msmith@xiph.org>
 ** Copyright (C) 2007 John ffitch
 **
@@ -68,7 +68,7 @@
 #include "sfendian.h"
 #include "common.h"
 
-#if HAVE_EXTERNAL_LIBS
+#if HAVE_EXTERNAL_XIPH_LIBS
 
 #include <ogg/ogg.h>
 #include <vorbis/codec.h>
@@ -76,12 +76,13 @@
 
 #include "ogg.h"
 
-typedef int convert_func (int, void *, int, int, float **) ;
+typedef int convert_func (SF_PRIVATE *psf, int, void *, int, int, float **) ;
 
 static int	vorbis_read_header (SF_PRIVATE *psf, int log_data) ;
 static int	vorbis_write_header (SF_PRIVATE *psf, int calc_length) ;
 static int	vorbis_close (SF_PRIVATE *psf) ;
 static int	vorbis_command (SF_PRIVATE *psf, int command, void *data, int datasize) ;
+static int	vorbis_byterate (SF_PRIVATE *psf) ;
 static sf_count_t	vorbis_seek (SF_PRIVATE *psf, int mode, sf_count_t offset) ;
 static sf_count_t	vorbis_read_s (SF_PRIVATE *psf, short *ptr, sf_count_t len) ;
 static sf_count_t	vorbis_read_i (SF_PRIVATE *psf, int *ptr, sf_count_t len) ;
@@ -108,6 +109,8 @@ static STR_PAIRS vorbis_metatypes [] =
 	{	SF_STR_DATE,		"Date" },
 	{	SF_STR_ALBUM,		"Album" },
 	{	SF_STR_LICENSE,		"License" },
+	{	SF_STR_TRACKNUMBER,	"Tracknumber" },
+	{	SF_STR_GENRE, 		"Genre" },
 } ;
 
 typedef struct
@@ -132,7 +135,7 @@ vorbis_read_header (SF_PRIVATE *psf, int log_data)
 	OGG_PRIVATE *odata = (OGG_PRIVATE *) psf->container_data ;
 	VORBIS_PRIVATE *vdata = (VORBIS_PRIVATE *) psf->codec_data ;
 	char *buffer ;
-	int	 bytes ;
+	int	bytes ;
 	int i, nn ;
 
 	odata->eos = 0 ;
@@ -249,7 +252,7 @@ vorbis_read_header (SF_PRIVATE *psf, int log_data)
 	**	header page is the only place where missing data is fatal.
 	*/
 
-	i = 0 ;			 /* Count of number of packets read */
+	i = 0 ;			/* Count of number of packets read */
 	while (i < 2)
 	{	int result = ogg_sync_pageout (&odata->osync, &odata->opage) ;
 		if (result == 0)
@@ -349,10 +352,10 @@ vorbis_write_header (SF_PRIVATE *psf, int UNUSED (calc_length))
 
 #if 0
 	ret = vorbis_encode_init (&vdata->vinfo, psf->sf.channels, psf->sf.samplerate, -1, 128000, -1) ; /* average bitrate mode */
-	ret = ( vorbis_encode_setup_managed (&vdata->vinfo, psf->sf.channels,
-						 psf->sf.samplerate, -1, 128000, -1) ||
-		vorbis_encode_ctl (&vdata->vinfo, OV_ECTL_RATEMANAGE_AVG, NULL) ||
-		vorbis_encode_setup_init (&vdata->vinfo)) ;
+	ret = (	vorbis_encode_setup_managed (&vdata->vinfo, psf->sf.channels, psf->sf.samplerate, -1, 128000, -1)
+			|| vorbis_encode_ctl (&vdata->vinfo, OV_ECTL_RATEMANAGE_AVG, NULL)
+			|| vorbis_encode_setup_init (&vdata->vinfo)
+			) ;
 #endif
 	if (ret)
 		return SFE_BAD_OPEN_FORMAT ;
@@ -366,22 +369,25 @@ vorbis_write_header (SF_PRIVATE *psf, int UNUSED (calc_length))
 	for (k = 0 ; k < SF_MAX_STRINGS ; k++)
 	{	const char * name ;
 
-		if (psf->strings [k].type == 0)
+		if (psf->strings.data [k].type == 0)
 			break ;
 
-		switch (psf->strings [k].type)
-		{	case SF_STR_TITLE :		name = "TITLE" ; break ;
-			case SF_STR_COPYRIGHT : name = "COPYRIGHT" ; break ;
-			case SF_STR_SOFTWARE :	name = "SOFTWARE" ; break ;
-			case SF_STR_ARTIST :	name = "ARTIST" ; break ;
-			case SF_STR_COMMENT :	name = "COMMENT" ; break ;
-			case SF_STR_DATE :		name = "DATE" ; break ;
-			case SF_STR_ALBUM :		name = "ALBUM" ; break ;
-			case SF_STR_LICENSE :	name = "LICENSE" ; break ;
+		switch (psf->strings.data [k].type)
+		{	case SF_STR_TITLE :			name = "TITLE" ; break ;
+			case SF_STR_COPYRIGHT : 	name = "COPYRIGHT" ; break ;
+			case SF_STR_SOFTWARE :		name = "SOFTWARE" ; break ;
+			case SF_STR_ARTIST :		name = "ARTIST" ; break ;
+			case SF_STR_COMMENT :		name = "COMMENT" ; break ;
+			case SF_STR_DATE :			name = "DATE" ; break ;
+			case SF_STR_ALBUM :			name = "ALBUM" ; break ;
+			case SF_STR_LICENSE :		name = "LICENSE" ; break ;
+			case SF_STR_TRACKNUMBER : 	name = "Tracknumber" ; break ;
+			case SF_STR_GENRE : 		name = "Genre" ; break ;
+
 			default : continue ;
 			} ;
 
-		vorbis_comment_add_tag (&vdata->vcomment, name, psf->strings [k].str) ;
+		vorbis_comment_add_tag (&vdata->vcomment, name, psf->strings.storage + psf->strings.data [k].offset) ;
 		} ;
 
 	/* set up the analysis state and auxiliary encoding storage */
@@ -483,7 +489,7 @@ vorbis_close (SF_PRIVATE *psf)
 int
 ogg_vorbis_open (SF_PRIVATE *psf)
 {	OGG_PRIVATE* odata = psf->container_data ;
-	VORBIS_PRIVATE* vdata = calloc (1, sizeof (VORBIS_PRIVATE)) ;
+	VORBIS_PRIVATE* vdata ;
 	int	error = 0 ;
 
 	if (odata == NULL)
@@ -491,6 +497,7 @@ ogg_vorbis_open (SF_PRIVATE *psf)
 		return SFE_INTERNAL ;
 		} ;
 
+	vdata = calloc (1, sizeof (VORBIS_PRIVATE)) ;
 	psf->codec_data = vdata ;
 
 	if (psf->file.mode == SFM_RDWR)
@@ -525,14 +532,12 @@ ogg_vorbis_open (SF_PRIVATE *psf)
 		psf->write_double	= vorbis_write_d ;
 
 		psf->sf.frames = SF_COUNT_MAX ; /* Unknown really */
-		psf->str_flags = SF_STR_ALLOW_START ;
+		psf->strings.flags = SF_STR_ALLOW_START ;
 		} ;
 
-	psf->bytewidth = 1 ;
-	psf->blockwidth = psf->bytewidth * psf->sf.channels ;
-
 	psf->seek = vorbis_seek ;
 	psf->command = vorbis_command ;
+	psf->byterate = vorbis_byterate ;
 
 	/* FIXME, FIXME, FIXME : Hack these here for now and correct later. */
 	psf->sf.format = SF_FORMAT_OGG | SF_FORMAT_VORBIS ;
@@ -550,14 +555,14 @@ vorbis_command (SF_PRIVATE *psf, int command, void * data, int datasize)
 {	VORBIS_PRIVATE *vdata = (VORBIS_PRIVATE *) psf->codec_data ;
 
 	switch (command)
-	{	case SFC_SET_VBR_ENCODING_QUALITY :
+	{	case SFC_SET_COMPRESSION_LEVEL :
 			if (data == NULL || datasize != sizeof (double))
 				return SF_FALSE ;
 
 			if (psf->have_written)
 				return SF_FALSE ;
 
-			vdata->quality = *((double *) data) ;
+			vdata->quality = 1.0 - *((double *) data) ;
 
 			/* Clip range. */
 			vdata->quality = SF_MAX (0.0, SF_MIN (1.0, vdata->quality)) ;
@@ -573,36 +578,56 @@ vorbis_command (SF_PRIVATE *psf, int command, void * data, int datasize)
 } /* vorbis_command */
 
 static int
-vorbis_rnull (int samples, void *UNUSED (vptr), int UNUSED (off) , int channels, float **UNUSED (pcm))
+vorbis_rnull (SF_PRIVATE *UNUSED (psf), int samples, void *UNUSED (vptr), int UNUSED (off) , int channels, float **UNUSED (pcm))
 {
 	return samples * channels ;
 } /* vorbis_rnull */
 
 static int
-vorbis_rshort (int samples, void *vptr, int off, int channels, float **pcm)
+vorbis_rshort (SF_PRIVATE *psf, int samples, void *vptr, int off, int channels, float **pcm)
 {
 	short *ptr = (short*) vptr + off ;
 	int i = 0, j, n ;
-	for (j = 0 ; j < samples ; j++)
-		for (n = 0 ; n < channels ; n++)
-			ptr [i++] = lrintf (pcm [n][j] * 32767.0f) ;
+	if (psf->float_int_mult)
+	{
+		float inverse = 1.0 / psf->float_max ;
+		for (j = 0 ; j < samples ; j++)
+			for (n = 0 ; n < channels ; n++)
+				ptr [i++] = lrintf ((pcm [n][j] * inverse) * 32767.0f) ;
+	}
+	else
+	{
+		for (j = 0 ; j < samples ; j++)
+			for (n = 0 ; n < channels ; n++)
+				ptr [i++] = lrintf (pcm [n][j] * 32767.0f) ;
+	}
 	return i ;
 } /* vorbis_rshort */
 
 static int
-vorbis_rint (int samples, void *vptr, int off, int channels, float **pcm)
+vorbis_rint (SF_PRIVATE *psf, int samples, void *vptr, int off, int channels, float **pcm)
 {
 	int *ptr = (int*) vptr + off ;
 	int i = 0, j, n ;
 
-	for (j = 0 ; j < samples ; j++)
-		for (n = 0 ; n < channels ; n++)
-			ptr [i++] = lrintf (pcm [n][j] * 2147483647.0f) ;
+	if (psf->float_int_mult)
+	{
+		float inverse = 1.0 / psf->float_max ;
+		for (j = 0 ; j < samples ; j++)
+			for (n = 0 ; n < channels ; n++)
+				ptr [i++] = lrintf ((pcm [n][j] * inverse) * 2147483647.0f) ;
+	}
+	else
+	{
+		for (j = 0 ; j < samples ; j++)
+			for (n = 0 ; n < channels ; n++)
+				ptr [i++] = lrintf (pcm [n][j] * 2147483647.0f) ;
+	}
 	return i ;
 } /* vorbis_rint */
 
 static int
-vorbis_rfloat (int samples, void *vptr, int off, int channels, float **pcm)
+vorbis_rfloat (SF_PRIVATE *UNUSED (psf), int samples, void *vptr, int off, int channels, float **pcm)
 {
 	float *ptr = (float*) vptr + off ;
 	int i = 0, j, n ;
@@ -613,7 +638,7 @@ vorbis_rfloat (int samples, void *vptr, int off, int channels, float **pcm)
 } /* vorbis_rfloat */
 
 static int
-vorbis_rdouble (int samples, void *vptr, int off, int channels, float **pcm)
+vorbis_rdouble (SF_PRIVATE *UNUSED (psf), int samples, void *vptr, int off, int channels, float **pcm)
 {
 	double *ptr = (double*) vptr + off ;
 	int i = 0, j, n ;
@@ -636,7 +661,7 @@ vorbis_read_sample (SF_PRIVATE *psf, void *ptr, sf_count_t lens, convert_func *t
 
 	while ((samples = vorbis_synthesis_pcmout (&vdata->vdsp, &pcm)) > 0)
 	{	if (samples > len) samples = len ;
-		i += transfn (samples, ptr, i, psf->sf.channels, pcm) ;
+		i += transfn (psf, samples, ptr, i, psf->sf.channels, pcm) ;
 		len -= samples ;
 		/* tell libvorbis how many samples we actually consumed */
 		vorbis_synthesis_read (&vdata->vdsp, samples) ;
@@ -644,7 +669,7 @@ vorbis_read_sample (SF_PRIVATE *psf, void *ptr, sf_count_t lens, convert_func *t
 		if (len == 0)
 			return i ; /* Is this necessary */
 	}
-	goto start0 ;		 /* Jump into the nasty nest */
+	goto start0 ;		/* Jump into the nasty nest */
 	while (len > 0 && !odata->eos)
 	{
 		while (len > 0 && !odata->eos)
@@ -670,15 +695,16 @@ vorbis_read_sample (SF_PRIVATE *psf, void *ptr, sf_count_t lens, convert_func *t
 					{	/* we have a packet.	Decode it */
 						if (vorbis_synthesis (&vdata->vblock, &odata->opacket) == 0) /* test for success! */
 							vorbis_synthesis_blockin (&vdata->vdsp, &vdata->vblock) ;
-		  /*
-		  **pcm is a multichannel float vector.	 In stereo, for
-		  example, pcm [0] is left, and pcm [1] is right.	 samples is
-		  the size of each channel.	 Convert the float values
-		  (-1.<=range<=1.) to whatever PCM format and write it out */
+						/*
+						** pcm is a multichannel float vector.	 In stereo, for
+						** example, pcm [0] is left, and pcm [1] is right.	 samples is
+						** the size of each channel.	 Convert the float values
+						** (-1.<=range<=1.) to whatever PCM format and write it out.
+						*/
 
 						while ((samples = vorbis_synthesis_pcmout (&vdata->vdsp, &pcm)) > 0)
-						{	if (samples>len) samples = len ;
-							i += transfn (samples, ptr, i, psf->sf.channels, pcm) ;
+						{	if (samples > len) samples = len ;
+							i += transfn (psf, samples, ptr, i, psf->sf.channels, pcm) ;
 							len -= samples ;
 							/* tell libvorbis how many samples we actually consumed */
 							vorbis_synthesis_read (&vdata->vdsp, samples) ;
@@ -875,6 +901,16 @@ vorbis_seek (SF_PRIVATE *psf, int UNUSED (mode), sf_count_t offset)
 	return 0 ;
 } /* vorbis_seek */
 
+
+static int
+vorbis_byterate (SF_PRIVATE *psf)
+{
+	if (psf->file.mode == SFM_READ)
+		return (psf->datalength * psf->sf.samplerate) / psf->sf.frames ;
+
+	return -1 ;
+} /* vorbis_byterate */
+
 /*==============================================================================
 **	Most of the following code was snipped from Mike Smith's ogginfo utility
 **	which is part of vorbis-tools.
@@ -1130,7 +1166,7 @@ vorbis_length (SF_PRIVATE *psf)
 	return length ;
 } /* vorbis_length */
 
-#else /* HAVE_EXTERNAL_LIBS */
+#else /* HAVE_EXTERNAL_XIPH_LIBS */
 
 int
 ogg_vorbis_open	(SF_PRIVATE *psf)
