@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2008-2010 Erik de Castro Lopo <erikd@mega-nerd.com>
+** Copyright (C) 2008-2016 Erik de Castro Lopo <erikd@mega-nerd.com>
 ** Copyright (C) 2009      Uli Franke <cls@nebadje.org>
 **
 ** This program is free software; you can redistribute it and/or modify
@@ -35,12 +35,14 @@
 #include	"sndfile.h"
 #include	"sfendian.h"
 #include	"common.h"
-#include	"wav_w64.h"
+#include	"wavlike.h"
 
 /*------------------------------------------------------------------------------
 ** Macros to handle big/little endian issues.
 */
 #define	RF64_MARKER		MAKE_MARKER ('R', 'F', '6', '4')
+#define	RIFF_MARKER		MAKE_MARKER ('R', 'I', 'F', 'F')
+#define	JUNK_MARKER		MAKE_MARKER ('J', 'U', 'N', 'K')
 #define	FFFF_MARKER		MAKE_MARKER (0xff, 0xff, 0xff, 0xff)
 #define	WAVE_MARKER		MAKE_MARKER ('W', 'A', 'V', 'E')
 #define	ds64_MARKER		MAKE_MARKER ('d', 's', '6', '4')
@@ -49,33 +51,44 @@
 #define	data_MARKER		MAKE_MARKER ('d', 'a', 't', 'a')
 
 #define bext_MARKER		MAKE_MARKER ('b', 'e', 'x', 't')
+#define cart_MARKER		MAKE_MARKER ('c', 'a', 'r', 't')
 #define OggS_MARKER		MAKE_MARKER ('O', 'g', 'g', 'S')
-#define wvpk_MARKER 	MAKE_MARKER ('w', 'v', 'p', 'k')
+#define wvpk_MARKER		MAKE_MARKER ('w', 'v', 'p', 'k')
+#define LIST_MARKER		MAKE_MARKER ('L', 'I', 'S', 'T')
 
-/*------------------------------------------------------------------------------
-** Typedefs.
+/*
+** The file size limit in bytes below which we can, if requested, write this
+** file as a RIFF/WAVE file.
 */
 
+#define RIFF_DOWNGRADE_BYTES	((sf_count_t) 0xffffffff)
+
 /*------------------------------------------------------------------------------
 ** Private static functions.
 */
 
 static int	rf64_read_header (SF_PRIVATE *psf, int *blockalign, int *framesperblock) ;
 static int	rf64_write_header (SF_PRIVATE *psf, int calc_length) ;
+static int	rf64_write_tailer (SF_PRIVATE *psf) ;
 static int	rf64_close (SF_PRIVATE *psf) ;
 static int	rf64_command (SF_PRIVATE *psf, int command, void * UNUSED (data), int datasize) ;
 
+static int rf64_set_chunk (SF_PRIVATE *psf, const SF_CHUNK_INFO * chunk_info) ;
+static SF_CHUNK_ITERATOR * rf64_next_chunk_iterator (SF_PRIVATE *psf, SF_CHUNK_ITERATOR * iterator) ;
+static int rf64_get_chunk_size (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info) ;
+static int rf64_get_chunk_data (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info) ;
+
 /*------------------------------------------------------------------------------
 ** Public function.
 */
 
 int
 rf64_open (SF_PRIVATE *psf)
-{	WAV_PRIVATE *wpriv ;
+{	WAVLIKE_PRIVATE *wpriv ;
 	int	subformat, error = 0 ;
 	int blockalign, framesperblock ;
 
-	if ((wpriv = calloc (1, sizeof (WAV_PRIVATE))) == NULL)
+	if ((wpriv = calloc (1, sizeof (WAVLIKE_PRIVATE))) == NULL)
 		return SFE_MALLOC_FAILED ;
 	psf->container_data = wpriv ;
 	wpriv->wavex_ambisonic = SF_AMBISONIC_NONE ;
@@ -83,10 +96,15 @@ rf64_open (SF_PRIVATE *psf)
 	/* All RF64 files are little endian. */
 	psf->endian = SF_ENDIAN_LITTLE ;
 
+	psf->strings.flags = SF_STR_ALLOW_START | SF_STR_ALLOW_END ;
 
 	if (psf->file.mode == SFM_READ || (psf->file.mode == SFM_RDWR && psf->filelength > 0))
 	{	if ((error = rf64_read_header (psf, &blockalign, &framesperblock)) != 0)
 			return error ;
+
+		psf->next_chunk_iterator = rf64_next_chunk_iterator ;
+		psf->get_chunk_size = rf64_get_chunk_size ;
+		psf->get_chunk_data = rf64_get_chunk_data ;
 		} ;
 
 	if ((psf->sf.format & SF_FORMAT_TYPEMASK) != SF_FORMAT_RF64)
@@ -104,6 +122,7 @@ rf64_open (SF_PRIVATE *psf)
 			return error ;
 
 		psf->write_header = rf64_write_header ;
+		psf->set_chunk = rf64_set_chunk ;
 		} ;
 
 	psf->container_close = rf64_close ;
@@ -148,18 +167,21 @@ enum
 {	HAVE_ds64	= 0x01,
 	HAVE_fmt	= 0x02,
 	HAVE_bext	= 0x04,
-	HAVE_data	= 0x08
+	HAVE_data	= 0x08,
+	HAVE_cart	= 0x10,
+	HAVE_PEAK	= 0x20,
+	HAVE_other	= 0x40
 } ;
 
 #define HAVE_CHUNK(CHUNK)	((parsestage & CHUNK) != 0)
 
 static int
 rf64_read_header (SF_PRIVATE *psf, int *blockalign, int *framesperblock)
-{	WAV_PRIVATE	*wpriv ;
+{	WAVLIKE_PRIVATE	*wpriv ;
 	WAV_FMT		*wav_fmt ;
-	sf_count_t riff_size = 0, sample_count = 0 ;
-	unsigned int size32, parsestage = 0 ;
-	int marker, marks [2], error, done = 0, format = 0 ;
+	sf_count_t riff_size = 0, frame_count = 0, ds64_datalength = 0 ;
+	uint32_t marks [2], marker, chunk_size, parsestage = 0 ;
+	int error, done = 0, format = 0 ;
 
 	if ((wpriv = psf->container_data) == NULL)
 		return SFE_INTERNAL ;
@@ -167,79 +189,146 @@ rf64_read_header (SF_PRIVATE *psf, int *blockalign, int *framesperblock)
 
 	/* Set position to start of file to begin reading header. */
 	psf_binheader_readf (psf, "pmmm", 0, &marker, marks, marks + 1) ;
-	if (marker != RF64_MARKER || marks [0] != FFFF_MARKER || marks [1] != WAVE_MARKER)
+	if (marker != RF64_MARKER || marks [1] != WAVE_MARKER)
 		return SFE_RF64_NOT_RF64 ;
 
-	psf_log_printf (psf, "%M\n  %M\n", RF64_MARKER, WAVE_MARKER) ;
+	if (marks [0] == FFFF_MARKER)
+		psf_log_printf (psf, "%M\n  %M\n", RF64_MARKER, WAVE_MARKER) ;
+	else
+		psf_log_printf (psf, "%M : 0x%x (should be 0xFFFFFFFF)\n  %M\n", RF64_MARKER, WAVE_MARKER) ;
 
-	while (! done)
-	{	psf_binheader_readf (psf, "em4", &marker, &size32) ;
+	while (NOT (done))
+	{
+		marker = chunk_size = 0 ;
+		psf_binheader_readf (psf, "em4", &marker, &chunk_size) ;
+
+		if (marker == 0)
+		{	sf_count_t pos = psf_ftell (psf) ;
+			psf_log_printf (psf, "Have 0 marker at position %D (0x%x).\n", pos, pos) ;
+			break ;
+			} ;
+
+		psf_store_read_chunk_u32 (&psf->rchunks, marker, psf_ftell (psf), chunk_size) ;
 
 		switch (marker)
 		{	case ds64_MARKER :
-					psf_log_printf (psf, "%M : %u\n", marker, size32) ;
+				if (parsestage & HAVE_ds64)
+				{	psf_log_printf (psf, "*** Second 'ds64' chunk?\n") ;
+					break ;
+					} ;
+
+				{	unsigned int table_len, bytesread ;
 
 					/* Read ds64 sizes (3 8-byte words). */
-					psf_binheader_readf (psf, "888", &riff_size, &psf->datalength, &sample_count) ;
-					psf_log_printf (psf, "  Riff size : %D\n  Data size : %D\n  Frames    : %D\n",
-											riff_size, psf->datalength, sample_count) ;
+					bytesread = psf_binheader_readf (psf, "888", &riff_size, &ds64_datalength, &frame_count) ;
 
 					/* Read table length. */
-					psf_binheader_readf (psf, "4", &size32) ;
-					psf_log_printf (psf, "  Table length : %u\n", size32) ;
+					bytesread += psf_binheader_readf (psf, "4", &table_len) ;
+					/* Skip table for now. (this was "table_len + 4", why?) */
+					bytesread += psf_binheader_readf (psf, "j", table_len) ;
 
-					/* Skip table for now. (this was "size32 + 4", why?) */
-					psf_binheader_readf (psf, "j", size32) ;
-					parsestage |= HAVE_ds64 ;
-					break ;
+					if (chunk_size == bytesread)
+						psf_log_printf (psf, "%M : %u\n", marker, chunk_size) ;
+					else if (chunk_size >= bytesread + 4)
+					{	unsigned int next ;
+						psf_binheader_readf (psf, "m", &next) ;
+						if (next == fmt_MARKER)
+						{	psf_log_printf (psf, "%M : %u (should be %u)\n", marker, chunk_size, bytesread) ;
+							psf_binheader_readf (psf, "j", -4) ;
+							}
+						else
+						{	psf_log_printf (psf, "%M : %u\n", marker, chunk_size) ;
+							psf_binheader_readf (psf, "j", chunk_size - bytesread - 4) ;
+							} ;
+						} ;
+
+					if (psf->filelength != riff_size + 8)
+						psf_log_printf (psf, "  Riff size : %D (should be %D)\n", riff_size, psf->filelength - 8) ;
+					else
+						psf_log_printf (psf, "  Riff size : %D\n", riff_size) ;
+
+					psf_log_printf (psf, "  Data size : %D\n", ds64_datalength) ;
+
+					psf_log_printf (psf, "  Frames    : %D\n", frame_count) ;
+					psf_log_printf (psf, "  Table length : %u\n", table_len) ;
+
+					} ;
+				parsestage |= HAVE_ds64 ;
+				break ;
 
 			case fmt_MARKER:
-					psf_log_printf (psf, "%M : %u\n", marker, size32) ;
-					if ((error = wav_w64_read_fmt_chunk (psf, size32)) != 0)
+					psf_log_printf (psf, "%M : %u\n", marker, chunk_size) ;
+					if ((error = wavlike_read_fmt_chunk (psf, chunk_size)) != 0)
 						return error ;
 					format = wav_fmt->format ;
 					parsestage |= HAVE_fmt ;
 					break ;
 
 			case bext_MARKER :
-					if ((error = wav_read_bext_chunk (psf, size32)) != 0)
+					if ((error = wavlike_read_bext_chunk (psf, chunk_size)) != 0)
 						return error ;
 					parsestage |= HAVE_bext ;
 					break ;
 
+			case cart_MARKER :
+					if ((error = wavlike_read_cart_chunk (psf, chunk_size)) != 0)
+						return error ;
+					parsestage |= HAVE_cart ;
+					break ;
+
+			case INFO_MARKER :
+			case LIST_MARKER :
+					if ((error = wavlike_subchunk_parse (psf, marker, chunk_size)) != 0)
+						return error ;
+					parsestage |= HAVE_other ;
+					break ;
+
+			case PEAK_MARKER :
+					if ((parsestage & (HAVE_ds64 | HAVE_fmt)) != (HAVE_ds64 | HAVE_fmt))
+						return SFE_RF64_PEAK_B4_FMT ;
+
+					parsestage |= HAVE_PEAK ;
+
+					psf_log_printf (psf, "%M : %u\n", marker, chunk_size) ;
+					if ((error = wavlike_read_peak_chunk (psf, chunk_size)) != 0)
+						return error ;
+					psf->peak_info->peak_loc = ((parsestage & HAVE_data) == 0) ? SF_PEAK_START : SF_PEAK_END ;
+					break ;
+
 			case data_MARKER :
 				/* see wav for more sophisticated parsing -> implement state machine with parsestage */
 
-				psf_log_printf (psf, "%M : %x\n", marker, size32) ;
-
-				/* Fetch data length only from data chunk when data chunk size is not set to -1 */
-				if ( size32 != 0xFFFFFFFF )
-					psf->datalength = size32 ;
-				else if ( ! HAVE_CHUNK (HAVE_ds64))
-				{	/* Assuming the ds64 chunk comes always before the data chunk. */
-					psf_log_printf (psf, "  *** Data length not specified (-1 in data chunk and no ds64 chunk).") ;
+				if (HAVE_CHUNK (HAVE_ds64))
+				{	if (chunk_size == 0xffffffff)
+						psf_log_printf (psf, "%M : 0x%x\n", marker, chunk_size) ;
+					else
+						psf_log_printf (psf, "%M : 0x%x (should be 0xffffffff\n", marker, chunk_size) ;
+					psf->datalength = ds64_datalength ;
+					}
+				else
+				{	if (chunk_size == 0xffffffff)
+					{	psf_log_printf (psf, "%M : 0x%x\n", marker, chunk_size) ;
+						psf_log_printf (psf, "  *** Data length not specified no 'ds64' chunk.\n") ;
+						}
+					else
+					{	psf_log_printf (psf, "%M : 0x%x\n**** Weird, RF64 file without a 'ds64' chunk and no valid 'data' size.\n", marker, chunk_size) ;
+						psf->datalength = chunk_size ;
+						} ;
 					} ;
 
 				psf->dataoffset = psf_ftell (psf) ;
 
 				if (psf->dataoffset > 0)
-				{	if (size32 == 0 && riff_size == 8 && psf->filelength > 44)
+				{	if (chunk_size == 0 && riff_size == 8 && psf->filelength > 44)
 					{	psf_log_printf (psf, "  *** Looks like a WAV file which wasn't closed properly. Fixing it.\n") ;
 						psf->datalength = psf->filelength - psf->dataoffset ;
 						} ;
 
-					if (psf->datalength > psf->filelength - psf->dataoffset)
-					{	psf_log_printf (psf, "  data : %D (should be %D)\n", psf->datalength, psf->filelength - psf->dataoffset) ;
-						psf->datalength = psf->filelength - psf->dataoffset ;
-						}
-					else
-						psf_log_printf (psf, "  data : %D\n", psf->datalength) ;
-
 					/* Only set dataend if there really is data at the end. */
 					if (psf->datalength + psf->dataoffset < psf->filelength)
 						psf->dataend = psf->datalength + psf->dataoffset ;
 
-					if (! psf->sf.seekable || psf->dataoffset < 0)
+					if (NOT (psf->sf.seekable) || psf->dataoffset < 0)
 						break ;
 
 					/* Seek past data and continue reading header. */
@@ -252,24 +341,36 @@ rf64_read_header (SF_PRIVATE *psf, int *blockalign, int *framesperblock)
 				break ;
 
 			default :
+					if (chunk_size >= 0xffff0000)
+					{	psf_log_printf (psf, "*** Unknown chunk marker (%X) at position %D with length %u. Exiting parser.\n", marker, psf_ftell (psf) - 8, chunk_size) ;
+						done = SF_TRUE ;
+						break ;
+						} ;
+
 					if (isprint ((marker >> 24) & 0xFF) && isprint ((marker >> 16) & 0xFF)
 						&& isprint ((marker >> 8) & 0xFF) && isprint (marker & 0xFF))
-					{	psf_log_printf (psf, "*** %M : %d (unknown marker)\n", marker, size32) ;
-						if (size32 < 8)
-							done = SF_TRUE ;
-						psf_binheader_readf (psf, "j", size32) ;
+					{	psf_log_printf (psf, "*** %M : %d (unknown marker)\n", marker, chunk_size) ;
+						psf_binheader_readf (psf, "j", chunk_size) ;
 						break ;
 						} ;
 					if (psf_ftell (psf) & 0x03)
-					{	psf_log_printf (psf, "  Unknown chunk marker at position %d. Resynching.\n", size32 - 4) ;
+					{	psf_log_printf (psf, "  Unknown chunk marker at position 0x%x. Resynching.\n", chunk_size - 4) ;
 						psf_binheader_readf (psf, "j", -3) ;
 						break ;
 						} ;
-					psf_log_printf (psf, "*** Unknown chunk marker (%X) at position %D. Exiting parser.\n", marker, psf_ftell (psf) - 4) ;
+					psf_log_printf (psf, "*** Unknown chunk marker (0x%X) at position 0x%X. Exiting parser.\n", marker, psf_ftell (psf) - 4) ;
 					done = SF_TRUE ;
-				break ;
+					break ;
 			} ;	/* switch (marker) */
 
+		/* The 'data' chunk, a chunk size of 0xffffffff means that the 'data' chunk size
+		** is actually given by the ds64_datalength field.
+		*/
+		if (marker != data_MARKER && chunk_size >= psf->filelength)
+		{	psf_log_printf (psf, "*** Chunk size %u > file length %D. Exiting parser.\n", chunk_size, psf->filelength) ;
+			break ;
+			} ;
+
 		if (psf_ftell (psf) >= psf->filelength - SIGNED_SIZEOF (marker))
 		{	psf_log_printf (psf, "End\n") ;
 			break ;
@@ -277,7 +378,13 @@ rf64_read_header (SF_PRIVATE *psf, int *blockalign, int *framesperblock)
 		} ;
 
 	if (psf->dataoffset <= 0)
-		return SFE_WAV_NO_DATA ;
+		return SFE_RF64_NO_DATA ;
+
+	if (psf->sf.channels < 1)
+		return SFE_CHANNEL_COUNT_ZERO ;
+
+	if (psf->sf.channels >= SF_MAX_CHANNELS)
+		return SFE_CHANNEL_COUNT ;
 
 	/* WAVs can be little or big endian */
 	psf->endian = psf->rwf_endian ;
@@ -304,12 +411,15 @@ rf64_read_header (SF_PRIVATE *psf, int *blockalign, int *framesperblock)
 			psf->sf.frames = psf->datalength / psf->blockwidth ;
 		} ;
 
+	if (frame_count != psf->sf.frames)
+		psf_log_printf (psf, "*** Calculated frame count %d does not match value from 'ds64' chunk of %d.\n", psf->sf.frames, frame_count) ;
+
 	switch (format)
 	{
 		case WAVE_FORMAT_EXTENSIBLE :
 
 			/* with WAVE_FORMAT_EXTENSIBLE the psf->sf.format field is already set. We just have to set the major to rf64 */
-			psf->sf.format = (psf->sf.format & ~SF_FORMAT_TYPEMASK ) | SF_FORMAT_RF64 ;
+			psf->sf.format = (psf->sf.format & ~SF_FORMAT_TYPEMASK) | SF_FORMAT_RF64 ;
 
 			if (psf->sf.format == (SF_FORMAT_WAVEX | SF_FORMAT_MS_ADPCM))
 			{	*blockalign = wav_fmt->msadpcm.blockalign ;
@@ -359,10 +469,8 @@ rf64_read_header (SF_PRIVATE *psf, int *blockalign, int *framesperblock)
 		default : return SFE_UNIMPLEMENTED ;
 		} ;
 
-	psf_log_printf (psf, "  FORMAT : %x\n", psf->sf.format) ;
-
 	if (wpriv->fmt_is_broken)
-		wav_w64_analyze (psf) ;
+		wavlike_analyze (psf) ;
 
 	/* Only set the format endian-ness if its non-standard big-endian. */
 	if (psf->endian == SF_ENDIAN_BIG)
@@ -406,8 +514,8 @@ static const EXT_SUBFORMAT MSGUID_SUBTYPE_AMBISONIC_B_FORMAT_IEEE_FLOAT =
 
 
 static int
-wavex_write_fmt_chunk (SF_PRIVATE *psf)
-{	WAV_PRIVATE	*wpriv ;
+rf64_write_fmt_chunk (SF_PRIVATE *psf)
+{	WAVLIKE_PRIVATE	*wpriv ;
 	int subformat, fmt_size ;
 
 	if ((wpriv = psf->container_data) == NULL)
@@ -493,35 +601,39 @@ wavex_write_fmt_chunk (SF_PRIVATE *psf)
 		case SF_FORMAT_PCM_16 :
 		case SF_FORMAT_PCM_24 :
 		case SF_FORMAT_PCM_32 :
-			wavex_write_guid (psf, wpriv->wavex_ambisonic == SF_AMBISONIC_NONE ?
+			wavlike_write_guid (psf, wpriv->wavex_ambisonic == SF_AMBISONIC_NONE ?
 						&MSGUID_SUBTYPE_PCM : &MSGUID_SUBTYPE_AMBISONIC_B_FORMAT_PCM) ;
 			break ;
 
 		case SF_FORMAT_FLOAT :
 		case SF_FORMAT_DOUBLE :
-			wavex_write_guid (psf, wpriv->wavex_ambisonic == SF_AMBISONIC_NONE ?
+			wavlike_write_guid (psf, wpriv->wavex_ambisonic == SF_AMBISONIC_NONE ?
 						&MSGUID_SUBTYPE_IEEE_FLOAT : &MSGUID_SUBTYPE_AMBISONIC_B_FORMAT_IEEE_FLOAT) ;
 			break ;
 
 		case SF_FORMAT_ULAW :
-			wavex_write_guid (psf, &MSGUID_SUBTYPE_MULAW) ;
+			wavlike_write_guid (psf, &MSGUID_SUBTYPE_MULAW) ;
 			break ;
 
 		case SF_FORMAT_ALAW :
-			wavex_write_guid (psf, &MSGUID_SUBTYPE_ALAW) ;
+			wavlike_write_guid (psf, &MSGUID_SUBTYPE_ALAW) ;
 			break ;
 
 		default : return SFE_UNIMPLEMENTED ;
 		} ;
 
 	return 0 ;
-} /* wavex_write_fmt_chunk */
+} /* rf64_write_fmt_chunk */
 
 
 static int
 rf64_write_header (SF_PRIVATE *psf, int calc_length)
 {	sf_count_t	current ;
-	int 		error = 0, has_data = SF_FALSE ;
+	int 		error = 0, has_data = SF_FALSE, add_fact_chunk = 0 ;
+	WAVLIKE_PRIVATE	*wpriv ;
+
+	if ((wpriv = psf->container_data) == NULL)
+		return SFE_INTERNAL ;
 
 	current = psf_ftell (psf) ;
 
@@ -530,7 +642,6 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 
 	if (calc_length)
 	{	psf->filelength = psf_get_filelen (psf) ;
-
 		psf->datalength = psf->filelength - psf->dataoffset ;
 
 		if (psf->dataend)
@@ -545,10 +656,16 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 	psf->headindex = 0 ;
 	psf_fseek (psf, 0, SEEK_SET) ;
 
-	psf_binheader_writef (psf, "em4m", RF64_MARKER, 0xffffffff, WAVE_MARKER) ;
-
-	/* Currently no table. */
-	psf_binheader_writef (psf, "m48884", ds64_MARKER, 32, psf->filelength, psf->datalength, psf->sf.frames, 0) ;
+	if (wpriv->rf64_downgrade && psf->filelength < RIFF_DOWNGRADE_BYTES)
+	{	psf_binheader_writef (psf, "etm8m", RIFF_MARKER, (psf->filelength < 8) ? 8 : psf->filelength - 8, WAVE_MARKER) ;
+		psf_binheader_writef (psf, "m4884", JUNK_MARKER, 20, 0, 0, 0, 0) ;
+		add_fact_chunk = 1 ;
+		}
+	else
+	{	psf_binheader_writef (psf, "em4m", RF64_MARKER, 0xffffffff, WAVE_MARKER) ;
+		/* Currently no table. */
+		psf_binheader_writef (psf, "m48884", ds64_MARKER, 28, psf->filelength - 8, psf->datalength, psf->sf.frames, 0) ;
+		} ;
 
 	/* WAVE and 'fmt ' markers. */
 	psf_binheader_writef (psf, "m", fmt_MARKER) ;
@@ -562,8 +679,10 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 
 		case SF_FORMAT_WAVEX :
 		case SF_FORMAT_RF64 :
-				if ((error = wavex_write_fmt_chunk (psf)) != 0)
+				if ((error = rf64_write_fmt_chunk (psf)) != 0)
 					return error ;
+				if (add_fact_chunk)
+					psf_binheader_writef (psf, "tm48", fact_MARKER, 4, psf->sf.frames) ;
 				break ;
 
 		default :
@@ -571,23 +690,23 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 		} ;
 
 	if (psf->broadcast_16k != NULL)
-		wav_write_bext_chunk (psf) ;
+		wavlike_write_bext_chunk (psf) ;
+
+	if (psf->cart_16k != NULL)
+		wavlike_write_cart_chunk (psf) ;
 
-#if 0
 	/* The LIST/INFO chunk. */
-	if (psf->str_flags & SF_STR_LOCATE_START)
-		wav_write_strings (psf, SF_STR_LOCATE_START) ;
+	if (psf->strings.flags & SF_STR_LOCATE_START)
+		wavlike_write_strings (psf, SF_STR_LOCATE_START) ;
 
 	if (psf->peak_info != NULL && psf->peak_info->peak_loc == SF_PEAK_START)
-	{	psf_binheader_writef (psf, "m4", PEAK_MARKER, WAV_PEAK_CHUNK_SIZE (psf->sf.channels)) ;
-		psf_binheader_writef (psf, "44", 1, time (NULL)) ;
-		for (k = 0 ; k < psf->sf.channels ; k++)
-			psf_binheader_writef (psf, "ft8", (float) psf->peak_info->peaks [k].value, psf->peak_info->peaks [k].position) ;
-		} ;
+		wavlike_write_peak_chunk (psf) ;
 
-//	if (psf->broadcast_info != NULL)
-//		wav_write_bext_chunk (psf) ;
+	/* Write custom headers. */
+	if (psf->wchunks.used > 0)
+		wavlike_write_custom_chunks (psf) ;
 
+#if 0
 	if (psf->instrument != NULL)
 	{	int		tmp ;
 		double	dtune = (double) (0x40000000) / 25.0 ;
@@ -605,7 +724,7 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 		{	int type ;
 
 			type = psf->instrument->loops [tmp].mode ;
-			type = (type == SF_LOOP_FORWARD ? 0 : type==SF_LOOP_BACKWARD ? 2 : type == SF_LOOP_ALTERNATING ? 1 : 32) ;
+			type = (type == SF_LOOP_FORWARD ? 0 : type == SF_LOOP_BACKWARD ? 2 : type == SF_LOOP_ALTERNATING ? 1 : 32) ;
 
 			psf_binheader_writef (psf, "44", tmp, type) ;
 			psf_binheader_writef (psf, "44", psf->instrument->loops [tmp].start, psf->instrument->loops [tmp].end) ;
@@ -613,15 +732,18 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 			} ;
 		} ;
 
+#endif
+
 	if (psf->headindex + 8 < psf->dataoffset)
 	{	/* Add PAD data if necessary. */
-		k = psf->dataoffset - 16 - psf->headindex ;
+		int k = psf->dataoffset - 16 - psf->headindex ;
 		psf_binheader_writef (psf, "m4z", PAD_MARKER, k, make_size_t (k)) ;
 		} ;
 
-#endif
-
-	psf_binheader_writef (psf, "m4", data_MARKER, 0xffffffff) ;
+	if (wpriv->rf64_downgrade && (psf->filelength < RIFF_DOWNGRADE_BYTES))
+		psf_binheader_writef (psf, "tm8", data_MARKER, psf->datalength) ;
+	else
+		psf_binheader_writef (psf, "m4", data_MARKER, 0xffffffff) ;
 
 	psf_fwrite (psf->header, psf->headindex, 1, psf) ;
 	if (psf->error)
@@ -634,7 +756,7 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 
 	psf->dataoffset = psf->headindex ;
 
-	if (! has_data)
+	if (NOT (has_data))
 		psf_fseek (psf, psf->dataoffset, SEEK_SET) ;
 	else if (current > 0)
 		psf_fseek (psf, current, SEEK_SET) ;
@@ -642,13 +764,42 @@ rf64_write_header (SF_PRIVATE *psf, int calc_length)
 	return psf->error ;
 } /* rf64_write_header */
 
+static int
+rf64_write_tailer (SF_PRIVATE *psf)
+{
+	/* Reset the current header buffer length to zero. */
+	psf->header [0] = 0 ;
+	psf->headindex = 0 ;
+
+	if (psf->bytewidth > 0 && psf->sf.seekable == SF_TRUE)
+	{	psf->datalength = psf->sf.frames * psf->bytewidth * psf->sf.channels ;
+		psf->dataend = psf->dataoffset + psf->datalength ;
+		} ;
+
+	if (psf->dataend > 0)
+		psf_fseek (psf, psf->dataend, SEEK_SET) ;
+	else
+		psf->dataend = psf_fseek (psf, 0, SEEK_END) ;
+
+	if (psf->dataend & 1)
+		psf_binheader_writef (psf, "z", 1) ;
+
+	if (psf->strings.flags & SF_STR_LOCATE_END)
+		wavlike_write_strings (psf, SF_STR_LOCATE_END) ;
+
+	/* Write the tailer. */
+	if (psf->headindex > 0)
+		psf_fwrite (psf->header, psf->headindex, 1, psf) ;
+
+	return 0 ;
+} /* rf64_write_tailer */
+
 static int
 rf64_close (SF_PRIVATE *psf)
 {
 	if (psf->file.mode == SFM_WRITE || psf->file.mode == SFM_RDWR)
-	{	// rf64_write_tailer (psf) ;
-
-		psf->write_header (psf, SF_TRUE) ;
+	{	rf64_write_tailer (psf) ;
+		rf64_write_header (psf, SF_TRUE) ;
 		} ;
 
 	return 0 ;
@@ -656,7 +807,7 @@ rf64_close (SF_PRIVATE *psf)
 
 static int
 rf64_command (SF_PRIVATE *psf, int command, void * UNUSED (data), int datasize)
-{	WAV_PRIVATE	*wpriv ;
+{	WAVLIKE_PRIVATE	*wpriv ;
 
 	if ((wpriv = psf->container_data) == NULL)
 		return SFE_INTERNAL ;
@@ -677,9 +828,14 @@ rf64_command (SF_PRIVATE *psf, int command, void * UNUSED (data), int datasize)
 			return wpriv->wavex_ambisonic ;
 
 		case SFC_SET_CHANNEL_MAP_INFO :
-			wpriv->wavex_channelmask = wavex_gen_channel_mask (psf->channel_map, psf->sf.channels) ;
+			wpriv->wavex_channelmask = wavlike_gen_channel_mask (psf->channel_map, psf->sf.channels) ;
 			return (wpriv->wavex_channelmask != 0) ;
 
+		case SFC_RF64_AUTO_DOWNGRADE :
+			if (psf->have_written == 0)
+				wpriv->rf64_downgrade = datasize ? SF_TRUE : SF_FALSE ;
+			return wpriv->rf64_downgrade ;
+
 		default :
 			break ;
 	} ;
@@ -687,3 +843,46 @@ rf64_command (SF_PRIVATE *psf, int command, void * UNUSED (data), int datasize)
 	return 0 ;
 } /* rf64_command */
 
+static int
+rf64_set_chunk (SF_PRIVATE *psf, const SF_CHUNK_INFO * chunk_info)
+{	return psf_save_write_chunk (&psf->wchunks, chunk_info) ;
+} /* rf64_set_chunk */
+
+static SF_CHUNK_ITERATOR *
+rf64_next_chunk_iterator (SF_PRIVATE *psf, SF_CHUNK_ITERATOR * iterator)
+{	return psf_next_chunk_iterator (&psf->rchunks, iterator) ;
+} /* rf64_next_chunk_iterator */
+
+static int
+rf64_get_chunk_size (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info)
+{	int indx ;
+
+	if ((indx = psf_find_read_chunk_iterator (&psf->rchunks, iterator)) < 0)
+		return SFE_UNKNOWN_CHUNK ;
+
+	chunk_info->datalen = psf->rchunks.chunks [indx].len ;
+
+	return SFE_NO_ERROR ;
+} /* rf64_get_chunk_size */
+
+static int
+rf64_get_chunk_data (SF_PRIVATE *psf, const SF_CHUNK_ITERATOR * iterator, SF_CHUNK_INFO * chunk_info)
+{	int indx ;
+	sf_count_t pos ;
+
+	if ((indx = psf_find_read_chunk_iterator (&psf->rchunks, iterator)) < 0)
+		return SFE_UNKNOWN_CHUNK ;
+
+	if (chunk_info->data == NULL)
+		return SFE_BAD_CHUNK_DATA_PTR ;
+
+	chunk_info->id_size = psf->rchunks.chunks [indx].id_size ;
+	memcpy (chunk_info->id, psf->rchunks.chunks [indx].id, sizeof (chunk_info->id) / sizeof (*chunk_info->id)) ;
+
+	pos = psf_ftell (psf) ;
+	psf_fseek (psf, psf->rchunks.chunks [indx].offset, SEEK_SET) ;
+	psf_fread (chunk_info->data, SF_MIN (chunk_info->datalen, psf->rchunks.chunks [indx].len), 1, psf) ;
+	psf_fseek (psf, pos, SEEK_SET) ;
+
+	return SFE_NO_ERROR ;
+} /* rf64_get_chunk_data */
